#include <cmath>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <vector>

using namespace std;

#include "main.h"
#include "include_glut.h"
#include "points3.h"

#include "world.h"
#include "simulate.h"
#include "landuse.h"

#include "rendering.h"
#include "textures.h"
#include "illustrations.h"

void draw_terrain();
void draw_water();
void draw_villagers();
void draw_roads();
void draw_miscellaneous();

namespace Option {
  bool display = true;
  bool simulation = true;
  bool illustrations = false;
  
  bool use_normals = true;
  bool use_textures = true;
  bool use_colors = true;

  bool blob_paths = false;
  bool blob_direction = false;

  bool render_grid = false;
  bool render_vectors = false;
  int render_canyons = 1; // 0, 1, or 2
}

namespace Rendering {
  int frame; // TODO: temporary
  int triangle_count;
}

namespace PickName {
  GLuint TERRAIN = 1;
  GLuint WATER = 2;
  GLuint BLOB = 3;
}

/* Mouse state -- move this into a class or namespace */
namespace Pointer {
  bool mouse_1_down = false;
  bool mouse_2_down = false;
  bool mouse_3_down = false;

  /* These are screen coordinates for the mouse pointer,
     with (0,0) in the lower left hand corner of the window */
  int screen_x = 0;
  int screen_y = 0;
}

int current_blob = -1;
int mouseM = 0;
int mouseN = 0;
int zoom_key_down = 0; // <0 for zoom in, >0 for zoom out
Point3d mouseLocation;
GLint rendering_mode = GL_RENDER;
enum { VIEW, CONSTRUCTION } annotation_type = VIEW;

/* Debugging */
ostream& operator << (ostream& out, Point3d const & p) {
  return out << "Point(" << p.x << ',' << p.y << ',' << p.z << ')';
}

/* Rendering of the game map */
namespace Rendering {
  // These will remain fixed for all time
  array x;
  array y;

  // TODO: the {z, nx, ny, nz} bundle should be a separate Surface object.
  
  // These will change whenever the base altitude changes
  array z_base;
  array r_base, g_base, b_base, a_base; // color
  
  // These will change whenever the base altitude OR the water level changes
  array z_water;
  array r_water, g_water, b_water, a_water; // color
  
  // This describes the area of the grid that can be seen.
  // frustrum_map[m] will give a pair of values, min_n:max_n.  We only
  // have to look at min_n <= n < max_n (half-open interval).  Many
  // values of m will have an empty range (where min_n >= max_n).
  int frustrum_map[Map::size][2];
}

/* Camera control */
namespace Camera {
  Point3d center_of_view(400, 400, 0);
  Point3d center_of_view_target(120, 50, 0.0);
  Vector3d angle_of_view(0.0, -1.0, 0.0);
  GLdouble mvmatrix[16], projmatrix[16];
  GLint viewport[4];

  /* Coordinate conversion from 2d screen (window) coordinates to 3d
     world coordinates */
  Point3d screen_to_world(float screenX, float screenY, float worldZ) {
    double wx0, wy0, wz0;
    gluUnProject(screenX, screenY, 0.0,
                 mvmatrix, projmatrix, viewport, &wx0, &wy0, &wz0);
    double wx1, wy1, wz1;
    gluUnProject(screenX, screenY, 1.0,
                 mvmatrix, projmatrix, viewport, &wx1, &wy1, &wz1);
    
    // Compute a position on the line (wx0,wy0,wz0)-(wx1,wy1,wz1)
    // where z = worldZ
    float depth = (worldZ-wz0) / (wz1-wz0);
    float x = wx0 + (wx1-wx0)*depth;
    float y = wy0 + (wy1-wy0)*depth;
    return Point3d(x, y, worldZ);
  }
  
  Point3d compute_mouse_position(float worldZ) {
    return screen_to_world(Pointer::screen_x, Pointer::screen_y, worldZ);
  }
}

void _draw_empty_vertex(int m, int n,
                        float*& vertices, float*& normals,
                        float*& colors,
                        float*& textures0, float*& textures1, float*& textures2) {
  if (Option::use_colors) {
    *colors++ = 0.0f;
    *colors++ = 0.0f;
    *colors++ = 0.0f;
    *colors++ = 1.0f;
  }
  if (Option::use_textures) {
    *textures0++ = 0.0f;
    *textures0++ = 0.0f;

    *textures1++ = 0.0f;
    *textures1++ = 0.0f;
    *textures1++ = 0.0f;
  }
  if (Option::use_normals) {
    *normals++ = 0.0f;
    *normals++ = 0.0f;
    *normals++ = 1.0f;
  }
  Point3d p(grid_to_world(m, n));
  *vertices++ = p.x;
  *vertices++ = p.y;
  *vertices++ = +15.0f;
}

void _draw_triangle_vertex(int m, int n, bool color, float alpha,
                           float*& vertices, float*& normals,
                           float*& colors,
                           float*& textures0, float*& textures1, float*& textures2) {
  using namespace Rendering;
  float const T = 0.07f*4.0f; // texture scale for landscape
  
  float px = x[m][n];
  float py = y[m][n];
  float pz = z_base[m][n];
  if (rendering_mode == GL_RENDER) {
    if (Option::use_textures) {
      // The three textures are rotated 45 degrees from one another to
      // avoid visual artifacts from textures lining up
      *textures0++ = px*T;
      *textures0++ = py*T;

      *textures1++ = (px-py)*0.013f;
      *textures1++ = (py+px)*0.013f;
      *textures1++ = pz*1.1f;

      *textures2++ = py*T*1.7f;
      *textures2++ = px*T*1.7f;
    }
    if (Option::use_colors) {
      *colors++ = r_base[m][n];
      *colors++ = g_base[m][n];
      *colors++ = b_base[m][n];
      *colors++ = min(alpha, a_base[m][n]);
    }
    if (Option::use_normals) {
      *normals++ = Map::nx_base[m][n];
      *normals++ = Map::ny_base[m][n];
      *normals++ = Map::nz_base[m][n];
    }
  }
  
  *vertices++ = px;
  *vertices++ = py;
  *vertices++ = pz;
}

void draw_triangle_strip(int m, int min_n, int max_n,
                         bool color, float alpha=1.0) {
  // TODO: If we use indexed geometry, we can store the vertices and
  // normals elsewhere, and just pass in the indices of the vertices
  // we actually want to draw.  It's not clear if this is going to be
  // faster in practice than triangle strips.

  // TODO: The NVidia vertex array range extension allows glDrawArrays
  // to return immediately, so that we can run some stuff on our CPU
  // while the GPU does its work.  (Maybe we can interleave simulation
  // and rendering??)  See http://www.levp.de/3d/ .  We'd have to
  // allocate all the vertices, colors, texture coordinates, and
  // normals in one big array (48 bytes per map location = 768k),
  // which lives in AGP memory (??).  AGP has better write performance
  // than read performance.  Or we can have two areas in the AGP
  // memory, and fill one while the GPU is rendering the other.  Also
  // see http://www.geocities.com/kenschwarz/opengl/superhack/ .. and
  // see
  // http://developer.nvidia.com/docs/IO/1377/ATT/GDC2K_ogl_extensions.pdf
  // for more about extensions.  And http://developer.nvidia.com/docs/IO/1323/ATT/VertexArrayRange.pdf is a really good doc.
  float vertices[2*3*(2+Map::size)];
  float* vertices_end = vertices;

  float normals[2*3*(2+Map::size)];
  float* normals_end = normals;

  float colors[2*4*(2+Map::size)];
  float* colors_end = colors;

  float textures0[2*3*(2+Map::size)];
  float* textures0_end = textures0;

  float textures1[2*3*(2+Map::size)];
  float* textures1_end = textures1;

  float textures2[2*2*(2+Map::size)];
  float* textures2_end = textures2;

  // We need to put this extra vertex in here to make sure we draw in
  // counter-clockwise order, so the normals are correct.
  _draw_triangle_vertex(m, min_n, color, alpha,
                        vertices_end, normals_end, colors_end,
                        textures0_end, textures1_end, textures2_end);

  int m1 = m + Hex::dm[1];
  int dn = Hex::dn[m&1][1];
  for (int n = min_n; n <= max_n; ++n) {
    int n1 = n + dn;
    if (Map::valid(m1, n1)) {
      _draw_triangle_vertex(m1, n1, color, alpha,
                            vertices_end, normals_end, colors_end,
                            textures0_end, textures1_end, textures2_end);
    } else {
      _draw_empty_vertex(m1, n1,
                         vertices_end, normals_end, colors_end,
                         textures0_end, textures1_end, textures2_end);
    }

    if (Map::valid(m, n)) {
      // Note: (m, n) where min_n <= n < max_n is valid, but we are
      // looping over n <= max_n in order to get the edge of the map
      // (or at least, part of it).  Once we have better handling of
      // the map edge, we can get rid of the redundant validity test.
      _draw_triangle_vertex(m, n, color, alpha,
                            vertices_end, normals_end, colors_end,
                            textures0_end, textures1_end, textures2_end);
    }
  }

  glVertexPointer(3, GL_FLOAT, 0, vertices);
  glEnableClientState(GL_VERTEX_ARRAY);

  if (normals_end != normals) {
    glNormalPointer(GL_FLOAT, 0, normals);
    glEnableClientState(GL_NORMAL_ARRAY);
  } else {
    glNormal3f(0.0f, 0.0f, 1.0f);
  }
  
  if (colors_end != colors) {
    glColorPointer(4, GL_FLOAT, 0, colors);
    glEnableClientState(GL_COLOR_ARRAY);
  } else {
    glColor4f(0.7f, 0.7f, 0.7f, alpha);
  }
  
  if (Option::use_textures && textures0_end != textures0) {
    glClientActiveTextureARB(GL_TEXTURE0_ARB);
    glTexCoordPointer(2, GL_FLOAT, 0, textures0);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTextureARB(GL_TEXTURE1_ARB);
    glTexCoordPointer(3, GL_FLOAT, 0, textures1);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTextureARB(GL_TEXTURE2_ARB);
    glTexCoordPointer(2, GL_FLOAT, 0, textures2);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  } else {
    glTexCoord2f(0.0f, 0.0f);
  }
  
  glDrawArrays(GL_TRIANGLE_STRIP, 0, (vertices_end-vertices)/3);
  Rendering::triangle_count += ((vertices_end-vertices)/3 - 2);
  
  glClientActiveTextureARB(GL_TEXTURE0_ARB);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  glClientActiveTextureARB(GL_TEXTURE1_ARB);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  glClientActiveTextureARB(GL_TEXTURE2_ARB);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  
  glDisableClientState(GL_COLOR_ARRAY);
  glDisableClientState(GL_NORMAL_ARRAY);
  glDisableClientState(GL_VERTEX_ARRAY);

  if (Option::render_vectors) {
    glBegin(GL_LINES);
    glNormal3f(0.0f, 0.0f, 1.0f);
    glTexCoord2f(0.0f, 0.0f);
    for (int n = min_n; n <= max_n; ++n) {
      if (Map::valid(m, n)) {
        glColor3f(1.0f, 0.0f, 0.0f);
        glVertex3f(Rendering::x[m][n],
                   Rendering::y[m][n],
                   Rendering::z_water[m][n]);
        glVertex3f(Map::nx_base[m][n] + Rendering::x[m][n],
                   Map::ny_base[m][n] + Rendering::y[m][n],
                   Map::nz_base[m][n] + Rendering::z_water[m][n]);
        glColor3f(0.0f, 1.0f, 0.0f);
        glVertex3f(Rendering::x[m][n],
                   Rendering::y[m][n],
                   Rendering::z_water[m][n]);
        glVertex3f(Map::gx_base[m][n] + Rendering::x[m][n],
                   Map::gy_base[m][n] + Rendering::y[m][n],
                   Map::gz_base[m][n] + Rendering::z_water[m][n]);
        glColor3f(0.0f, 0.5f, 1.0f);
        glVertex3f(Rendering::x[m][n],
                   Rendering::y[m][n],
                   Rendering::z_water[m][n]);
        glVertex3f(Map::water_momentum[m][n].x + Rendering::x[m][n],
                   Map::water_momentum[m][n].y + Rendering::y[m][n],
                   Rendering::z_water[m][n]);
      }
    }
    glEnd();
  }
}

namespace Rendering {

  void _draw_primitive(
    Point3d const & p1, float r1, float g1, float b1, float a1,
    Point3d const & p2, float r2, float g2, float b2, float a2,
    Point3d const & p3, float r3, float g3, float b3, float a3,
    float tx, float ty) {

    float const texture_coord_multiplier = 8.0;
    tx *= texture_coord_multiplier;
    ty *= texture_coord_multiplier;
    float const T0 = 0.1*texture_coord_multiplier; // texture scale for water
    float const T1 = 0.0012*texture_coord_multiplier;
    float const T1_dx = 0.0003f;
    float const T1_dy = 0.0005f;
    float const T2_dx = 0.0004f;
    float const T2_dy = 0.0001f;
    
    glBegin(GL_TRIANGLES);
    glColor4f(r1, g1, b1, a1);
    glMultiTexCoord2fARB(GL_TEXTURE0_ARB, p1.x*T0+tx, p1.y*T0+ty);
    glMultiTexCoord2fARB(GL_TEXTURE1_ARB, p1.y*T1+frame*T1_dx, p1.x*T1+frame*T1_dy);
    glMultiTexCoord2fARB(GL_TEXTURE2_ARB, p1.x*T1+frame*T2_dx, p1.y*T1+frame*T2_dy);
    glVertex3f(p1.x, p1.y, p1.z);
    glColor4f(r2, g2, b2, a2);
    glMultiTexCoord2fARB(GL_TEXTURE0_ARB, p2.x*T0+tx, p2.y*T0+ty);
    glMultiTexCoord2fARB(GL_TEXTURE1_ARB, p2.y*T1+frame*T1_dx, p2.x*T1+frame*T1_dy);
    glMultiTexCoord2fARB(GL_TEXTURE2_ARB, p2.x*T1+frame*T2_dx, p2.y*T1+frame*T2_dy);
    glVertex3f(p2.x, p2.y, p2.z);
    glColor4f(r3, g3, b3, a3);
    glMultiTexCoord2fARB(GL_TEXTURE0_ARB, p3.x*T0+tx, p3.y*T0+ty);
    glMultiTexCoord2fARB(GL_TEXTURE1_ARB, p3.y*T1+frame*T1_dx, p3.x*T1+frame*T1_dy);
    glMultiTexCoord2fARB(GL_TEXTURE2_ARB, p3.x*T1+frame*T2_dx, p3.y*T1+frame*T2_dy);
    glVertex3f(p3.x, p3.y, p3.z);
    glEnd();
    Rendering::triangle_count += 1;
    
    if (Option::render_grid) {
      glBegin(GL_LINES);
      glColor4f(r1*0.5, g1*0.5, b1*0.5, 1);
      glVertex3f(p1.x, p1.y, p1.z);
      glColor4f(r2*0.5, g2*0.5, b2*0.5, 1.0f);
      glVertex3f(p2.x, p2.y, p2.z);
      glColor4f(r3*0.5, g3*0.5, b3*0.5, 1.0f);
      glVertex3f(p3.x, p3.y, p3.z);
      glEnd();
    }
  }

  int wet_neighbors(int m, int n) {
    // Count how many neighbors have water
    int count = 0;
    for (int k = 1; k <= 6; k++) {
      int m2 = m + Hex::dm[k], n2 = n + Hex::dn[m&1][k];
      count += Map::valid(m2, n2) && Map::water[m2][n2] > 0;
    }
    return count;
  }

  void draw_water_triangles(int m1, int n1, int k, array z,
                          array r, array g, array b, array a, float alpha) {
    // First compute the three coordinates that we are going to consider
    // NOTE: we are assuming that k is 1 or 2 so we don't have to mod 6
    int m2 = m1 + Hex::dm[k],   n2 = n1 + Hex::dn[m1&1][k];
    int m3 = m1 + Hex::dm[k+1], n3 = n1 + Hex::dn[m1&1][k+1];

    // TODO: remove this limitation :( First, we need to compute the
    // hexagon centers for hexagons outside the normal range.  Then we
    // need to compute colors by using a hexagon color instead of a
    // midpoint.
    if (!(Map::valid(m1, n1) || Map::valid(m2, n2) || Map::valid(m3, n3))) {
      return;
    }
    
    // Decide whether each small triangle is to be drawn
    bool t1 = Map::valid(m1, n1) && Map::water[m1][n1] > 0;
    bool t2 = Map::valid(m2, n2) && Map::water[m2][n2] > 0;
    bool t3 = Map::valid(m3, n3) && Map::water[m3][n3] > 0;

    if (!(t1 || t2 || t3)) {
      // Small optimization
      return;
    }

    // The hexagon centers
    Point3d p1(x[m1][n1], y[m1][n1], z[m1][n1]);
    Point3d p2(x[m2][n2], y[m2][n2], z[m2][n2]);
    Point3d p3(x[m3][n3], y[m3][n3], z[m3][n3]);
    float r1 = r[m1][n1], g1 = g[m1][n1], b1 = b[m1][n1], a1 = a[m1][n1];
    float r2 = r[m2][n2], g2 = g[m2][n2], b2 = b[m2][n2], a2 = a[m2][n2];
    float r3 = r[m3][n3], g3 = g[m3][n3], b3 = b[m3][n3], a3 = a[m3][n3];
    
    // Lower all points that are not watery.  This is very simple, and
    // produces its own set of artifacts, but it's fast and the other
    // methods I have so far aren't much better.  The actual code for
    // lowering z_water is in recompute_water_points().

    // Keep a texture offset per triangle to be drawn
    static float x_offset[Map::size][Map::size][2];
    static float y_offset[Map::size][Map::size][2];
    static bool offsets_initialized = false;
    if (!offsets_initialized) {
      // Set all the offsets to be random to try to avoid obvious grid appearance
      for (int m = 0; m < Map::size; ++m) {
        for (int n = 0; n < Map::size; ++n) {
          x_offset[m][n][0] = 0.001f * (random() % 1000);
          x_offset[m][n][1] = 0.001f * (random() % 1000);
          y_offset[m][n][0] = 0.001f * (random() % 1000);
          y_offset[m][n][1] = 0.001f * (random() % 1000);
        }
      }
      offsets_initialized = true;
    }
    
    if (rendering_mode == GL_RENDER) {
      // This will have to do until I figure out whether it's really
      // the right thing to do.  Velocity = momentum/mass so maybe
      // I should divide by the water value
      x_offset[m1][n1][k] -= 0.01f * (Map::water_momentum[m1][n1].x + 
                                      Map::water_momentum[m2][n2].x +
                                      Map::water_momentum[m3][n3].x);
      y_offset[m1][n1][k] -= 0.01f * (Map::water_momentum[m1][n1].y +
                                      Map::water_momentum[m2][n2].y +
                                      Map::water_momentum[m3][n3].y);
    }

    /* We're using this texture offset to represent the water
       direction visually.  The speed of the texture across the
       hex will be proportional to the volume of water that
       travels.  This isn't quite realistic because subsurface
       water flow shouldn't affect the texture speed, but it is
       close enough and it looks cool. */
    float tx = x_offset[m1][n1][k], ty = y_offset[m1][n1][k];

    if (Option::render_canyons >= 1) {
      // a1 = a2 = a3 = alpha;
    } else {
      a1 *= alpha;
      a2 *= alpha;
      a3 *= alpha;
    }
    _draw_primitive(p1, r1, g1, b1, a1,
                    p2, r2, g2, b2, a2,
                    p3, r3, g3, b3, a3,
                    tx, ty);
  }
}

float eyeHeight = 25.0;

struct GameApp: GlApp {

  GameApp() {
    cout << "Generating a map." << endl;
    setup_map();

    for (int m = 0; m < Map::size; ++m) {
      Rendering::frustrum_map[m][0] = 0;
      Rendering::frustrum_map[m][1] = Map::size;
    }
    recompute_base_points();
    recompute_base_colors();
    recompute_water_points();
  }
  
  bool set_flag(string const & flag) {
    if (flag == "no-textures") Option::use_textures = false;
    else if (flag == "white") Option::use_colors = false;
    else if (flag == "render-grid") Option::render_grid = true;
    else if (flag == "no-display") Option::display = false;
    else if (flag == "no-simulation") Option::simulation = false;
    else if (flag == "illustration") {
      Option::illustrations = true;
      eyeHeight = 100.0f;
    }
    else return GlApp::set_flag(flag);

    return true;
  }

  void setup_map() {
    srandom(time(NULL));

    LOOP(m, n) {
      Point3d p = grid_to_world(m, n);
      Rendering::x[m][n] = p.x;
      Rendering::y[m][n] = p.y;
    }
    
    for (int k = 0; k < 7; ++k) {
      Vector3d v = grid_to_world(Hex::dm[k], Hex::dn[0][k]) - grid_to_world(0, 0);
      Hex::neighbor_dx[k] = v.x;
      Hex::neighbor_dy[k] = v.y;
    }

    Map::setup_map();
    Map::setup_land_use();
  }

  void recompute_base_colors();
  void recompute_base_points();
  void recompute_water_points();

  virtual void paint_annotation() {
    /* Annotate with the game score.  */
    GlApp::paint_annotation();
    glPushMatrix();
    glTranslatef(0.0, 25.0, 0.0);
    glLineWidth(2.0);
    glScalef(0.2, 0.2, 0.2);
    glColor3f(0.0, 1.0, 0.8);
    char text[64];
    if (frame < 500) {
      sprintf(text, "Initial flooding");
    } else {
      sprintf(text, "Simulation %s, %d tri/frame",
              Option::simulation? "on":"off",
              Rendering::triangle_count);
      Rendering::triangle_count = 0;
      // TODO: fix this
      // sprintf(text, "Green score: %.02f%%", 100.0*(total_vegetation/Map::size/Map::size));
    }
    for (char const * p = text; *p; ++p) {
      glutStrokeCharacter(GLUT_STROKE_ROMAN, *p);
    }
    glLineWidth(1.0);
    glPopMatrix();
  }
  
  virtual void setup_view() {
    /* Change view if the mouse button is down */
    using namespace Camera;
    if (Pointer::mouse_3_down) {
      center_of_view_target = compute_mouse_position(0.0);
    }
    center_of_view += 0.03*(center_of_view_target - center_of_view);
    Vector3d n = center_of_view_target - center_of_view;
    if (n.length() > 0.1) {
      angle_of_view = normalize(angle_of_view + 0.0008*n);
    }

    /* Change zoom level */
    if (zoom_key_down < 0) {
      Option::zoom = min(200.0f, Option::zoom * 1.01f);
    } else if (zoom_key_down > 0) {
      Option::zoom = max(20.0f, Option::zoom / 1.01f);
    }
    
    /* Set up matrices */
#if 0
    // HACK: first person view, using the first villager (needs work)
    Point3d p(Map::villagers[0].x, Map::villagers[0].y, 0.0f);
    Point3d q(p.x + Map::villagers[0].dx, p.y + Map::villagers[0].dy, 0.0f);
    p.z = world_to_z(p.x, p.y);
    p.x -= (q.x - p.x);
    p.y -= (q.y - p.y);
    q.z = p.z - 7.0f;
    gluLookAt(p.x, p.y, p.z,
              q.x, q.y, q.z,
              0.0, 0.0, 1.0);
#else
    // TODO: as we're rotating the camera, also change the up angle so
    // that it looks like we're turning like a plane/helicopter
    Point3d eye = center_of_view + -25*angle_of_view + Vector3d(0, 0, eyeHeight);
    // Adjust the eye so that it's always 'zoom' away from center_of_view.
    float eye_distance = (eye-center_of_view).length();
    eye = center_of_view - ((center_of_view - eye) * (Option::zoom/eye_distance));
    gluLookAt(eye.x, eye.y, eye.z,
              center_of_view.x, center_of_view.y, center_of_view.z,
              0.0, 0.0, 1.0);// UP angle
#endif
    
    glGetIntegerv(GL_VIEWPORT, viewport);
    glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix);
    glGetDoublev(GL_PROJECTION_MATRIX, projmatrix);

    /* The back of a polygon can just be lines, not filled .. */
    glPolygonMode(GL_BACK, GL_LINE);
    
    /* If I draw multiple things at the same depth, I want to
     overwrite.  This isn't useful in most applications, but here we
     want to be able to layer effects on the map.

     See http://www.frii.com/~martz/oglfaq/polygonoffset.htm
    */
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_POLYGON_OFFSET_FILL); // use glPolygonOffset
  }

  virtual void update_selection() {
    // Optimization: if the camera hasn't changed and the mouse hasn't
    // changed, don't bother updating the selection.  This isn't quite
    // what we want, because even if the camera hasn't moved, the
    // objects in the world may have moved.  So we still run this
    // every 10th frame.

    if (frame % 10 != 0) {
      static Point3d camera_center;
      static int mouse_x = 0;
      static int mouse_y = 0;
      if (camera_center.x == Camera::center_of_view.x
          && camera_center.y == Camera::center_of_view.y
          && Pointer::screen_x == mouse_x && Pointer::screen_y == mouse_y) {
        return;
      }
      camera_center = Camera::center_of_view;
      mouse_x = Pointer::screen_x;
      mouse_y = Pointer::screen_y;
    }
    
    GLuint selectionBuffer[512];
    glSelectBuffer(512, selectionBuffer);
    glRenderMode(GL_SELECT);
    
    glInitNames();

    /* 3d perspective */
    glViewport(0, 0, window_width, window_height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPickMatrix(Pointer::screen_x, Pointer::screen_y,
                  1, 1, Camera::viewport);
    setup_projection();
    glMatrixMode(GL_MODELVIEW);
    
    rendering_mode = GL_SELECT;
    paint_content();
    rendering_mode = GL_RENDER;

    glFlush();

    // Find out how many hits we got.  -1 means too many to fit in our array
    GLint num_hits = glRenderMode(GL_RENDER);
    GLuint best_depth = GLuint(-1); // max_uint
    current_blob = -1;
    for (GLint pos = 0, i = 0; i < num_hits; ++i) {
      GLuint num_names = selectionBuffer[pos++];
      GLuint min_depth = selectionBuffer[pos++];
      ++pos; // GLuint max_depth = selectionBuffer[pos++];
      if (num_names > 0) {
        GLuint hit_type = selectionBuffer[pos];
        
        if (num_names == 3 &&
            (hit_type == PickName::TERRAIN || hit_type == PickName::WATER)) {
          if (min_depth < best_depth) {
            best_depth = min_depth;
            mouseM = selectionBuffer[pos+1];
            mouseN = selectionBuffer[pos+2];
          }
        } else if (hit_type == PickName::BLOB) {
          assert(num_names == 2);
          current_blob = selectionBuffer[pos+1];
        } else {
          /*
          cout << "Warning! Unknown hit: ";
          for (unsigned j = 0; j < num_names; ++j) {
            cout << selectionBuffer[pos+j];
          }
          cout << endl;
          */
        }
        pos += num_names;
      }
    }
  }
    
  void determine_viewable_map_area() {
    using namespace Camera;

    // Find the z range for column m of the map
    float z_min = -15.0f, z_max = +15.0f;
    static float last_z = 0.0f;

    if (Map::valid(mouseM, mouseN)) {
      last_z = Map::height[mouseM][mouseN];
    }
    
    // What are the screen coordinates of the viewable area?
    float x1 = viewport[0], x2 = viewport[2];
    float y1 = viewport[1] - 10, y2 = viewport[3] + 10;
    if (rendering_mode != GL_RENDER) {
      x1 = Pointer::screen_x - 20;
      x2 = Pointer::screen_x + 20;
      y1 = Pointer::screen_y - 100;
      y2 = Pointer::screen_y + 100;
      z_min = last_z - 5.0f;
      z_max = last_z + 5.0f;
    }
    // First find the four points of two trapezoids: one at z_min,
    // one at z_max.  Since we divide by the change in x between
    // two adjacent vertices, we need to make adjacent vertices have
    // different x.  This isn't quite right; we should be taking the
    // convex hull of the z_min and z_max trapezoids...
    Point3d v[12] = {
      screen_to_world(x1, y1, z_min),
      screen_to_world(x2, y1, z_min),
      screen_to_world(x2+1, y2, z_min),
      screen_to_world(x1-1, y2, z_min),
      screen_to_world(x1, y1, z_max),
      screen_to_world(x2, y1, z_max),
      screen_to_world(x2+1, y2, z_max),
      screen_to_world(x1-1, y2, z_max),
    };
      
    // For each m value, find the n value range that is viewable
    for (int m = 0; m < Map::size; ++m) {
      // We want to look at both sides of the hexagon, so we'll
      // compute the min/max y for x over (m +/- 0.9)*xWidth.  We
      // use 0.9 because we want the full hexagon width, not just
      // the offset from one column to the next.  (TODO: figure out
      // exactly how much this should be.)
      
      // TODO: Is this the best loop structure?  Maybe we can
      // compute multiple t values and not have the x_offset loop at
      // all.
      double min_y = +1e6, max_y = -1e6;
      double const offset_width = 0.9;
      for (double x_offset = -offset_width-0.1;
           x_offset <= offset_width; x_offset += offset_width) {
        double x = (m+x_offset)*Hex::xWidth;
        for (unsigned i = 0; i < sizeof(v)/sizeof(v[0]); ++i) {
          unsigned j = (i%4 == 3)? (i-3) : (i+1);
          double t = (x - v[i].x)/(double(v[j].x) - double(v[i].x));
          if (-0.01 <= t && t <= 1.01) {
            double y = v[i].y + t * (v[j].y - v[i].y);
            min_y = min(min_y, y);
            max_y = max(max_y, y);
          }
        }
      }
      // In addition to these n values, we want to get the ones with
      // centers just off the edge of the viewable area, because
      // their vertices might be inside the viewable area.
      float x = m*Hex::xWidth;
      int min_n = world_to_grid(x, min_y).second;
      int max_n = world_to_grid(x, max_y).second;

      // Note: the min:max range is half-open
      Rendering::frustrum_map[m][0] = max(0, min_n-3);
      Rendering::frustrum_map[m][1] = min(Map::size, max_n+4);
      // We also have to update column (m-1) because of the way
      // triangle strips are rendered (when drawing strip m, it draws
      // the area between column m and column m+1).
      if (m > 0) {
        Rendering::frustrum_map[m-1][0] = min(
          Rendering::frustrum_map[m-1][0],
          Rendering::frustrum_map[m][0]);
        Rendering::frustrum_map[m-1][1] = min(
          Rendering::frustrum_map[m-1][1],
          Rendering::frustrum_map[m][1]);
      }
    }
  }
    
  virtual void paint_content() {
    using namespace Camera;

    Rendering::frame = GlApp::frame;
    if (rendering_mode == GL_RENDER) {
      if (frame == 0) {
        glutSetWindowTitle("SimBlob 2 - development testbed");
        glutSetIconTitle("SimBlob 2");
        Texture::setup();
      }
    }

    if (Option::illustrations) {
      Illustrations::hexagonal_grid();
      return;
    }
    
    // HACK: use last mouse Z value.  TODO: only use this when the mouse hasn't
    // moved much
    mouseLocation = compute_mouse_position(
      Map::valid(mouseM, mouseN)?
      (Map::height[mouseM][mouseN]+Map::water[mouseM][mouseN]):0.0);

    // Another HACK: set the point light source to be at the mouse
    extern Point3d point_light_source_location;
    point_light_source_location = mouseLocation;
    point_light_source_location.z = 25.0f;
    
    if (current_blob != -1) {
      glutSetCursor(GLUT_CURSOR_INFO);
    } else if (Map::valid(mouseM, mouseN)) {
      glutSetCursor(GLUT_CURSOR_NONE);
    } else {
      glutSetCursor(GLUT_CURSOR_CROSSHAIR);
    }
    
    determine_viewable_map_area();
    
    if (rendering_mode == GL_RENDER) {
      /* Simulation */
      if (Option::simulation) {
        Map::simulate();
        Map::simulate_villagers();
      }

      if (Option::display) {
        /* React to user input */
        alter_terrain();

        if (Option::simulation) {
          if (frame % 30 == 2 || annotation_type == CONSTRUCTION) {
            recompute_base_points();
          }
      
          if (frame % 3 == 0) {
            recompute_base_colors();
          }
          
          recompute_water_points();
        }
      }
    }

    if (Option::display) {
      /* Draw everything */
      glPolygonOffset(0.0, -2.0);
      draw_terrain();
      glPolygonOffset(0.0, 0.0);
      draw_water();
      glPolygonOffset(-0.05, 0.0);
      
      // TODO: need to draw the sides of the map (from water down to a
      // very low point)

      draw_roads();
      draw_miscellaneous();
      draw_villagers();
      
      if (rendering_mode == GL_RENDER) {
        /* Everything else being drawn won't have shading */
        glNormal3f(0.0, 0.0, 1.0);
        draw_mouse_pointer(mouseLocation);
      }
    }

    if (rendering_mode == GL_RENDER) {
      /* Annotate with the spring locations */
      for (vector<Map::WaterSource>::iterator i = Map::water_sources.begin();
           i != Map::water_sources.end(); ++i) {
        glPushMatrix();
        Point3d p = grid_to_world(i->m, i->n,
                                Rendering::z_water[i->m][i->n] + 1.0f);
        glTranslatef(p.x, p.y, p.z);
        glLineWidth(2.0);
        glScalef(0.01, 0.01, 0.01);
        glRotatef(180.0, 0.0, 0.0, 1.0);
        glColor3f(0.5, 1.0, 1.0);
        for (char const * p = "* Spring"; *p; ++p) {
          glutStrokeCharacter(GLUT_STROKE_ROMAN, *p);
        }
        glPopMatrix();
        glLineWidth(1.0);
      }
    }
  }

  void draw_mouse_pointer(Point3d mouse) {
    glColor3f(1.0, 1.0, 1.0);
    glLineWidth(1.0);
    glEnable(GL_LINE_SMOOTH);
    glBegin(GL_LINES);

    /* Main vertical stem */
    glVertex3f(mouse.x, mouse.y, -30.0);
    glVertex3f(mouse.x, mouse.y, +30.0);

    /* Tick marks */
    for (int height = -10; height < 10; ++height) {
      float tick_length = 0.1 * ((height == 0)? 10 : (height % 5 == 0)? 5 : 2);
      
      glVertex3f(mouse.x - tick_length, mouse.y, height);
      glVertex3f(mouse.x + tick_length, mouse.y, height);
      glVertex3f(mouse.x, mouse.y - tick_length, height);
      glVertex3f(mouse.x, mouse.y + tick_length, height);
    }

    glEnd();

    // Highlight the current hexagon
    glNormal3f(0.0, 0.0, 1.0);
    glTexCoord2f(0.0, 0.0);
    glColor4f(3.0, 0.0, 0.0, 1.0);
    // Rendering::draw_solid_hexagon(mouseM, mouseN, Rendering::z_base);
    Rendering::draw_outline_hexagon(mouseM, mouseN, Rendering::z_base);
    
    glDisable(GL_LINE_SMOOTH);
    
  }
  
  virtual void event_mouse_motion(int x, int y) {
    using namespace Pointer;
    using namespace Camera;
    static bool outOfRange = false;

    screen_x = x;
    screen_y = (viewport[3] - 1) - y;

    if (screen_x < viewport[0] || screen_y < viewport[1] ||
        screen_x >= viewport[2] || screen_y >= viewport[3] ) {
      if (!outOfRange) {
        cout << "Mouse of out range: " << x << ',' << y << endl;
        outOfRange = true;
      }
      /* Note: even if it's out of range, it can still be captured by
       a mouse drag.  When dragging an object, this calls for
       auto-scrolling.  */
    } else {
      outOfRange = false;
    }
    
    if (Pointer::mouse_1_down) {
      // Do something with left drag
    }
  }

  virtual void event_mouse_button(int button, int state, int x, int y) {
    event_mouse_motion(x, y);
    if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
      cout << "Grabbing mouse " << button << endl;
      Pointer::mouse_1_down = true;
      annotation_type = CONSTRUCTION;
    } else if (button == GLUT_LEFT_BUTTON && state == GLUT_UP) {
      // cout << "Releasing mouse " << button << endl;
      Pointer::mouse_1_down = false;
      annotation_type = VIEW;
    } else if (button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN) {
      Pointer::mouse_2_down = true;
      annotation_type = CONSTRUCTION;
    } else if (button == GLUT_MIDDLE_BUTTON && state == GLUT_UP) {
      Pointer::mouse_2_down = false;
      annotation_type = VIEW;
    } else if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN) {
      Pointer::mouse_3_down = true;
    } else if (button == GLUT_RIGHT_BUTTON && state == GLUT_UP) {
      Pointer::mouse_3_down = false;
    }
  }

  virtual void event_key(unsigned char key, int x, int y, bool down) {
    cout << "key " << (down?"down ":"up ")
         << int(key) << " @ " << x << "," << y << endl;

    if (down && key == ' ') {
      Camera::center_of_view = grid_to_world(Map::size/2, Map::size/2, 0.0);
      Camera::center_of_view_target = Camera::center_of_view;
      Camera::angle_of_view = Vector3d(0.0, -1.0, 0.0);
    }
    if (down && key == 's') {
      Option::simulation = !Option::simulation;
      cout << "Simulation = " << Option::simulation << endl;
    }
    if (down && key == 'g') {
      Option::render_grid = !Option::render_grid;
      cout << "Grid rendering option = " << Option::render_grid << endl;
    }
    if (down && key == 't') {
      Option::use_textures = !Option::use_textures;
      cout << "Texturing option = " << Option::use_textures << endl;
    }
    if (down && key == 'w') {
      Option::use_colors = !Option::use_colors;
      cout << "White/gray rendering = " << (!Option::use_colors) << endl;
    }
    if (down && key == 'v') {
      Option::render_vectors = !Option::render_vectors;
      cout << "Vector field = " << Option::render_vectors << endl;
    }
    if (down && key == 'c') {
      Option::render_canyons = (1 + Option::render_canyons)%3;
      cout << "Canyon rendering option = " << Option::render_canyons << endl;
    }
    if (down && key == 'f') {
      Map::flooding = !Map::flooding;
      cout << "Flooding = " << Map::flooding << endl;
    }
    if (down && key == 'b') {
      Option::blob_paths = !Option::blob_paths;
      Option::blob_direction = !Option::blob_direction; // HACK
      cout << "Blob paths = " << Option::blob_paths << endl;
    }
    if (down && key == 'p') {
      Option::perspective = !Option::perspective;
      cout << "Perspective = " << Option::perspective << endl;
    }
    if (key == '+' || key == '=') {
      if (down) {
        ++zoom_key_down;
        cout << "Zooming in" << endl;
      } else {
        --zoom_key_down;
      }
    }
    if (key == '-') {
      if (down) {
        --zoom_key_down;
        cout << "Zooming out" << endl;
      } else {
        ++zoom_key_down;
      }
    }
    if (key == 17 /* Ctrl-Q */ ) {
      exit(0);
    }
  }

  virtual void event_special_key(int key, int x, int y, bool down) {
    cout << "special key " << (down? "down ":"up ")
         << int(key) << " @ " << x << "," << y << endl;

    if (down && key == GLUT_KEY_PAGE_DOWN) {
      eyeHeight = 25.0;
    } else if (down && key == GLUT_KEY_PAGE_UP) {
      eyeHeight = 75.0;
    }
    
    // Keys are GLUT_KEY_F{1-12}, GLUT_KEY_{LEFT,RIGHT,UP,DOWN},
    // GLUT_KEY_PAGE_{UP,DOWN}, GLUT_KEY_{HOME,END,INSERT}

    // To get keyboard modifiers, glutGetModifiers() is a bit set
    // GLUT_ACTIVE_SHIFT|GLUT_ACTIVE_CTRL_GLUT_ACTIVE_ALT
  }
  
  void alter_terrain() {
    if (Map::valid(mouseM, mouseN)) {
      if (Pointer::mouse_1_down) {
        Map::height[mouseM][mouseN] += 0.15;
        Map::vegetation[mouseM][mouseN] *= 0.95;
        Map::moisture[mouseM][mouseN] *= 0.95;
      }
      if (Pointer::mouse_2_down) {
        Map::height[mouseM][mouseN] -= 0.15;
        Map::vegetation[mouseM][mouseN] *= 0.99;
        Map::moisture[mouseM][mouseN] *= 0.99;
      }
    }
  }
};

// Draw a tree; assume we are inside GL_TRIANGLES 
void draw_tree(const Point3d& p, float dx, float dy, float tree_size) {
  // Trunk
  glColor4f(0.5, 0.4, 0.2, 1.0);
  glTexCoord2f(-1, 0);
  glVertex3f(p.x-dx*0.3*tree_size, p.y-dy*0.3*tree_size, p.z+0.3*tree_size);
  glTexCoord2f(1, 0);
  glVertex3f(p.x+dx*0.3*tree_size, p.y+dy*0.3*tree_size, p.z+0.3*tree_size);
  glTexCoord2f(0, 1);
  glVertex3f(p.x, p.y, p.z-0.2*tree_size);

  // Green
          
  // TODO: use GL_TRIANGLE_FAN?
          
  glColor4f(0.1, 0.5, 0.1, 1.0);
  glTexCoord2f(-1, 0);
  glVertex3f(p.x-dx*tree_size, p.y-dy*tree_size, p.z+0.3*tree_size);
  glColor4f(0.1, 0.5, 0.1, 1.0);
  glTexCoord2f(0, 1);
  glVertex3f(p.x, p.y, p.z+4*tree_size);
  glColor4f(0.2, 0.7, 0.2, 1.0);
  glTexCoord2f(0, 0);
  glVertex3f(p.x, p.y, p.z+0.3*tree_size);

  glColor4f(0.2, 0.7, 0.2, 1.0);
  glTexCoord2f(0, 0);
  glVertex3f(p.x, p.y, p.z+0.3*tree_size);
  glColor4f(0.1, 0.5, 0.1, 1.0);
  glTexCoord2f(0, 1);
  glVertex3f(p.x, p.y, p.z+4*tree_size);
  glColor4f(0.1, 0.3, 0.1, 1.0);
  glTexCoord2f(1, 0);
  glVertex3f(p.x+dx*tree_size, p.y+dy*tree_size, p.z+0.3*tree_size);
}

void draw_miscellaneous() {
  /* set up boring material */
  GLfloat material_diffuse[] = { 1, 1, 1, 1 };
  GLfloat material_specular[] = { 0, 0, 0, 1 };
  glMaterialfv(GL_FRONT, GL_DIFFUSE, material_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, material_specular);
  glMaterialf(GL_FRONT, GL_SHININESS, 0.0f);
  // TODO: abstract out the materials so we can reuse them in draw_*()
  
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, Texture::smooth);
  // Simple billboarding: orient the trees towards the camera
  float dx = -Camera::angle_of_view.y;
  float dy = +Camera::angle_of_view.x;
  glBegin(GL_TRIANGLES);
  glNormal3f(0.0, 0.0, 1.0);
  for (int m = 0; m < Map::size; ++m) {
    int min_n = Rendering::frustrum_map[m][0];
    int max_n = Rendering::frustrum_map[m][1];
    if (min_n >= max_n) continue;

    for (int n = min_n; n <= max_n; ++n) {
      if (Map::land_use[m][n].type != Map::TREE) continue;
      
      Point3d p(Rendering::x[m][n],
                Rendering::y[m][n],
                Rendering::z_base[m][n]);

      float tree_size = min(1.0f, 0.5f*Map::vegetation[m][n]);
      if (tree_size <= 0.1f) continue;

      draw_tree(p, dx, dy, tree_size);
    }
  }
  glEnd();
  glDisable(GL_TEXTURE_2D);
}

void draw_box(Point3d p1, Point3d p2) {
  // ASSUME we're in GL_QUADS

  float x1 = min(p1.x, p2.x), x2 = max(p1.x, p2.x);
  float y1 = min(p1.y, p2.y), y2 = max(p1.y, p2.y);
  float z1 = min(p1.z, p2.z), z2 = max(p1.z, p2.z);
  
  // Bottom
  glNormal3f(0.0f, 0.0f, -1.0f);
  glVertex3f(x1, y2, z1);
  glVertex3f(x2, y2, z1);
  glVertex3f(x2, y1, z1);
  glVertex3f(x1, y1, z1);

  // Top
  glNormal3f(0.0f, 0.0f, +1.0f);
  glVertex3f(x1, y1, z2);
  glVertex3f(x2, y1, z2);
  glVertex3f(x2, y2, z2);
  glVertex3f(x1, y2, z2);

  // Left
  glNormal3f(-1.0f, 0.0f, 0.0f);
  glVertex3f(x1, y1, z1);
  glVertex3f(x1, y1, z2);
  glVertex3f(x1, y2, z2);
  glVertex3f(x1, y2, z1);

  // Right
  glNormal3f(+1.0f, 0.0f, 0.0f);
  glVertex3f(x2, y2, z1);
  glVertex3f(x2, y2, z2);
  glVertex3f(x2, y1, z2);
  glVertex3f(x2, y1, z1);

  // Back
  glNormal3f(0.0f, -1.0f, 0.0f);
  glVertex3f(x1, y1, z1);
  glVertex3f(x2, y1, z1);
  glVertex3f(x2, y1, z2);
  glVertex3f(x1, y1, z2);
  
  // Front
  glNormal3f(0.0f, +1.0f, 0.0f);
  glVertex3f(x1, y2, z2);
  glVertex3f(x2, y2, z2);
  glVertex3f(x2, y2, z1);
  glVertex3f(x1, y2, z1);
}

// This is a pretty simple road.  It goes straight across the map and
// has a few supports.  It does not really exist in the simulation.
// It is drawn above the existing terrain.
void draw_roads() {
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, Texture::road);
  glBegin(GL_TRIANGLE_STRIP);
  glNormal3f(0.0, 0.0, 1.0);
  glColor3f(0.5, 0.5, 0.5);

  int m = 25;

  // Compute the draping
  float z[Map::size];
  for (int n = 0; n < Map::size; ++n) {
    z[n] = 0.0f;
  }
  
  const float max_slope = 0.3f;
  for (int iteration = 0; iteration != 5; iteration++) {
  for (int n = 3; n < Map::size-2; ++n) {
    float z1 = Rendering::z_base[m][n];
    float z2 = Rendering::z_base[m+1][n];
    z[n] = max(z[n], max(z1, z2));
    for (int n1 = n-20; n1 <= n+20; ++n1) {
      if (n1 > 0 && n1 < Map::size) {
        z[n] = max(z[n], z[n1]-abs(n-n1)*max_slope);
      }
    }
  }
  }
  
  // Draw the roadway
  float s = 0.0f, t = 0.0f;
  for (int n = 3; n < Map::size-2; ++n) {
    float x1 = Rendering::x[m][n], y1 = Rendering::y[m][n];
    float x2 = Rendering::x[m+1][n], y2 = Rendering::y[m+1][n];
    glTexCoord2f(s, 0.0); glVertex3f(x1, y1, z[n]);
    glTexCoord2f(s+(x2-x1), 1.0); glVertex3f(x2, y2, z[n]);
    s += Rendering::y[m][n+1] - Rendering::y[m][n];
  }
  glEnd();
  glDisable(GL_TEXTURE_2D);

  // Draw the supports
  glBegin(GL_QUADS);
  glColor3f(0.8, 0.8, 0.8);
  for (int n = 3; n < Map::size-2; ++n) {
    float x1 = Rendering::x[m][n], y1 = Rendering::y[m][n], z1 = Rendering::z_base[m][n];
    float x2 = Rendering::x[m+1][n], y2 = Rendering::y[m+1][n], z2 = Rendering::z_base[m+1][n];

    float z_low = min(z1, z2);
    draw_box(Point3d(x1+0.1f, y1, z_low), Point3d(x1+0.3f, y1+0.2f, z[n]-0.05f));
    draw_box(Point3d(x2-0.3f, y2, z_low), Point3d(x2-0.1f, y2+0.2f, z[n]-0.05f));
  }
  glEnd();
}

void draw_villagers() {
  // Now draw the villagers
  const float villager_radius = 0.4f;
  glPushName(PickName::BLOB);
  for (int i = 0; i < Map::NUM_VILLAGERS; ++i) {
    glPushName(i);
    Map::Villager& V = Map::villagers[i];
    Point3d p(V.x, V.y, 0.0f);
    p.z = world_to_z(p.x, p.y);

    glBegin(GL_TRIANGLE_FAN);
    /* NOTE: it's generally recommended that state changes (e.g.,
       glColor) occur outside the glBegin/glEnd section.  However, on
       my ATI card, it seems that sometimes the glColor doesn't
       actually take effect when I put it outside.  I suspect it's
       caching something incorrectly. So I've put the color inside the
       glBegin. */
    if (i == current_blob) {
      glColor3f(1.0f, 1.0f, 0.0f);
    } else {
      glColor3f(1.0f, 0.0f, 0.0f);
    }
    glNormal3f(0.0f, 0.0f, -1.0f);
    glVertex3f(p.x, p.y, p.z);
    glNormal3f(0.0f, -0.6f, 0.4f);
    glVertex3f(p.x+villager_radius, p.y-villager_radius, p.z+1.0f);
    glNormal3f(-0.6f, 0.0f, 0.4f);
    glVertex3f(p.x-villager_radius, p.y-villager_radius, p.z+1.0f);
    glNormal3f(0.0f, +0.6f, 0.4f);
    glVertex3f(p.x-villager_radius, p.y+villager_radius, p.z+1.0f);
    glNormal3f(+0.6f, 0.0f, 0.4f);
    glVertex3f(p.x+villager_radius, p.y+villager_radius, p.z+1.0f);
    glNormal3f(0.0f, -0.6f, 0.4f);
    glVertex3f(p.x+villager_radius, p.y-villager_radius, p.z+1.0f);
    glEnd();

    if (V.water > 0.0f) {
      float full = min(1.0f, V.water / 0.5f);
      glColor3f(0.0f, 0.0f, 1.0f);
      glNormal3f(0.0f, 0.0f, 1.0f);
      glBegin(GL_QUADS);
      glVertex3f(p.x+villager_radius*full, p.y-villager_radius*full, p.z+1.0f*full);
      glVertex3f(p.x+villager_radius*full, p.y+villager_radius*full, p.z+1.0f*full);
      glVertex3f(p.x-villager_radius*full, p.y+villager_radius*full, p.z+1.0f*full);
      glVertex3f(p.x-villager_radius*full, p.y-villager_radius*full, p.z+1.0f*full);
      glEnd();
    }
    
    glBegin(GL_LINES);
    glNormal3f(0.0f, 0.0f, 1.0f);
    if (i == current_blob || Option::blob_direction) {
      glColor3f(0.5f, 1.0f, 0.0f);
      glVertex3f(p.x, p.y, p.z);
      glVertex3f(p.x+V.dx, p.y+V.dy, p.z);
    }
    if (i == current_blob || Option::blob_paths) {
      glColor3f(0.0f, 0.0f, 0.0f);
      for (int j = 0; j < Map::VILLAGER_HISTORY-1; ++j) {
        glVertex3f(V.old_pos[j].x, V.old_pos[j].y, V.old_pos[j].z);
        glVertex3f(V.old_pos[j+1].x, V.old_pos[j+1].y, V.old_pos[j+1].z);
      }
    }
    glEnd();
    
    // HACK: we shouldn't be setting these here...
    V.old_pos[0].z = p.z;

    Vector3d vh(world_to_gradient(p.x, p.y, Map::height));
    V.dx += 0.003f * vh.x;
    V.dy += 0.003f * vh.y;

    // The gradient points towards LESS moisture
    Vector3d vm(world_to_gradient(p.x, p.y, Map::moisture));
    if (V.water > 0.0f) {
      // If we have water, move towards less moisture
      V.dx += 0.005f * vm.x;
      V.dy += 0.005f * vm.y;
    } else {
      // We have no water; move towards more moisture
      V.dx -= 0.005f * vm.x;
      V.dy -= 0.005f * vm.y;
    }

    glPopName();
  }

  glPopName();
  glNormal3f(0.0f, 0.0f, 1.0f);
}

// Linear interpolation.  Mix parameters together according to p.
static inline float mix(float x, float y, float p) {
  return x * (1.0f-p) + y * p;
}

void GameApp::recompute_base_colors() {
  for (int m0 = 0; m0 < Map::size; ++m0) {
    for (int n0 = Rendering::frustrum_map[m0][0];
         n0 < Rendering::frustrum_map[m0][1]; ++n0) {

      // Rules:
      // 1. The snow variable makes things white.
      // 2. Snow only occurs at high altitudes on flat land.
      // 3. The base land is gray at high altitudes.
      // 4. The vegetation variable makes things green on flat or medium slope land.
      // 4. Otherwise the land is brown.

      float flat_land = Map::nz_base[m0][n0]; // When the normal is up, z=1.0
      
      float vegetation = min(Map::vegetation[m0][n0], 2.0f*flat_land);
      float altitude_graying = max(0.0f, min(1.0f, (Map::height[m0][n0]-3.0f)*0.1f));
      float snow = min(1.0f - vegetation, min(altitude_graying, flat_land));

      // We apply the rules backwards. Base land, brown:
      float r = 0.4f, g = 0.4f, b = 0.2f;

      // Vegetation:
      r = mix(r, 0.4f, vegetation);
      g = mix(g, 0.8f, vegetation);
      b = mix(b, 0.2f, vegetation);
      
      // Gray at high altitudes:
      r = mix(r, 0.7f, altitude_graying);
      g = mix(g, 0.7f, altitude_graying);
      b = mix(b, 0.7f, altitude_graying);

      // Snow:
      r = mix(r, 0.95f, snow);
      g = mix(g, 0.95f, snow);
      b = mix(b, 0.95f, snow);

      Rendering::r_base[m0][n0] = r;
      Rendering::g_base[m0][n0] = g;
      Rendering::b_base[m0][n0] = b;
      Rendering::a_base[m0][n0] = 1.0f;

      /*
      // HACK: for checking height
      Rendering::r_base[m0][n0] = (Map::height[m0][n0] > 0.0)? 1.0:0.0;
      Rendering::g_base[m0][n0] = (m0 % 3 == 0)? 1.0:0.0;
      Rendering::b_base[m0][n0] = (Map::height[m0][n0] > 0.0)? 0.0:1.0;
      */
    }
  }
}

// Compute the derived information from the height field: normals and gradients
void GameApp::recompute_base_points() {
  for (int m0 = 0; m0 < Map::size; ++m0) {
    for (int n0 = Rendering::frustrum_map[m0][0];
         n0 < Rendering::frustrum_map[m0][1]; ++n0) {

      float z0 = Map::height[m0][n0];
      Point3d p0 = grid_to_world(m0, n0, z0);
      Vector3d central_normal(0.0f, 0.0f, 0.0f);
      Vector3d central_gradient(0.0f, 0.0f, 0.0f);
      
      Rendering::z_base[m0][n0] = z0;

      for (int k = 1; k < 7; ++k) {
        // We want to compute the normals of each face -- the hexagon
        // center, the k-neighbor, and its (k+1)-neighbor.

        // First neighbor
        int m1 = m0 + Hex::dm[k];
        int n1 = n0 + Hex::dn[m0&1][k];
        float z1 = Map::valid(m1, n1)? Map::height[m1][n1] : Map::height[m0][n0];
        Point3d p1 = grid_to_world(m1, n1, z1);

        // Second neighbor
        int kk = (k == 0)? 0 : ((k == 6)? 1 : k+1);
        int m2 = m0 + Hex::dm[kk];
        int n2 = n0 + Hex::dn[m0&1][kk];
        float z2 = Map::valid(m2, n2)? Map::height[m2][n2] : Map::height[m0][n0];
        Point3d p2 = grid_to_world(m2, n2, z2);

        Vector3d cross_product = cross(p1-p0, p2-p0);
        central_normal += cross_product;
        central_gradient += (z0-z1) * (p1-p0);
      }

      central_normal.z *= 0.3; // Enhance the lighting
      central_normal = normalize(central_normal);
      Map::nx_base[m0][n0] = central_normal.x;
      Map::ny_base[m0][n0] = central_normal.y;
      Map::nz_base[m0][n0] = central_normal.z;
      central_gradient = normalize(central_gradient);
      Map::gx_base[m0][n0] = central_gradient.x;
      Map::gy_base[m0][n0] = central_gradient.y;
      Map::gz_base[m0][n0] = central_gradient.z;
    }
  }
}

void GameApp::recompute_water_points() {
  for (int m0 = 0; m0 < Map::size; ++m0) {
    for (int n0 = Rendering::frustrum_map[m0][0];
         n0 < Rendering::frustrum_map[m0][1]; ++n0) {

      float w0 = Map::water[m0][n0];
      float z0 = Map::height[m0][n0] + w0;
      
      Rendering::z_water[m0][n0] = z0;

      if (w0 == 0.0) {
        Rendering::a_water[m0][n0] = 0.0;
        Rendering::r_water[m0][n0] = 0.1;
        Rendering::g_water[m0][n0] = 0.8;
        Rendering::b_water[m0][n0] = 1.0;

        if (Option::render_canyons >= 1) {
          for (int k = 1; k <= 6; ++k) {
            int m1 = m0 + Hex::dm[k];
            int n1 = n0 + Hex::dn[m0&1][k];
            float z1 = Map::valid(m1, n1)?
                       (Map::height[m1][n1] + Map::water[m1][n1])
                       : 0.0;
            if (z1 < z0) {
              z0 = z1;
            }
          }
        }
        float const alpha = 0.9f;
        Rendering::z_water[m0][n0] = alpha*z0 +
                                     (1.0f-alpha)*Map::height[m0][n0];
        
      } else {
        // Find the lowest non-watery neighbor
        if (Option::render_canyons >= 2) {
          z0 = 999.0f;
          for (int k = 1; k <= 6; ++k) {
            int m1 = m0 + Hex::dm[k];
            int n1 = n0 + Hex::dn[m0&1][k];
            if (Map::valid(m1, n1) && Map::water[m1][n1] == 0.0f) {
              float z1 = Map::height[m1][n1];
              if (z1 < z0) {
                z0 = z1;
              }
            }
          }
          if (Rendering::z_water[m0][n0] < z0 && z0 < 999.0f) {
            // Use most of the calculated value 
            z0 = 0.9*z0 + 0.1*Rendering::z_water[m0][n0];
            
            // We're in a canyon.  Pretend the canyon is simple and the
            // volume of the water roughly fits into a triangular wedge
            // -- thus volume water grows with the square of the water
            // height, instead of linearly (when we're not in a canyon).
            // If you solve the equations for this, you come up with:
            Rendering::z_water[m0][n0] = Map::height[m0][n0] +
                                         sqrt(2*Map::water[m0][n0]*
                                              (z0-Map::height[m0][n0]));
          }
        }
        
        float silt_coloring = min(1.0f, Map::silt[m0][n0]/(0.01f + w0));
        Rendering::r_water[m0][n0] = 5.0f/(50.0f+w0) * (1.0f-silt_coloring) + 0.6f*silt_coloring;
        Rendering::g_water[m0][n0] = 0.8f/(1.0f+w0) * (1.0f-silt_coloring) + 0.55f*silt_coloring;
        Rendering::b_water[m0][n0] = 10.0f/(10.0f+w0) * (1.0f-silt_coloring) + 0.4f*silt_coloring;
        
        Rendering::a_water[m0][n0] = min(1.0f, 0.6f+Map::water[m0][n0]);
      }
    }
  }
}

void draw_terrain() {
  /* set up boring material */
  GLfloat material_diffuse[] = { 1, 1, 1, 1 };
  GLfloat material_specular[] = { 0, 0, 0, 1 };
  glMaterialfv(GL_FRONT, GL_DIFFUSE, material_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, material_specular);
  glMaterialf(GL_FRONT, GL_SHININESS, 0.0f);

  if (Option::use_textures) {
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glBindTexture(GL_TEXTURE_2D, Texture::smooth);
    glEnable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glBindTexture(GL_TEXTURE_3D, Texture::contour);
    glEnable(GL_TEXTURE_3D);
    glActiveTextureARB(GL_TEXTURE2_ARB);
    glBindTexture(GL_TEXTURE_2D, Texture::smooth);
    glEnable(GL_TEXTURE_2D);
  }
  glPushName(PickName::TERRAIN);
  for (int m = 0; m < Map::size; ++m) {
    int min_n = Rendering::frustrum_map[m][0];
    int max_n = Rendering::frustrum_map[m][1];
    if (min_n >= max_n) continue;
    
    if (rendering_mode == GL_RENDER) {
      draw_triangle_strip(m, min_n, max_n, true);
    } else {
      for (int n = min_n; n < max_n; ++n) {
        Rendering::draw_solid_hexagon(m, n, Rendering::z_base);
      }
    }
  }
  glPopName();
  if (Option::use_textures) {
    glActiveTextureARB(GL_TEXTURE2_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glDisable(GL_TEXTURE_3D);
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glDisable(GL_TEXTURE_3D);
    glDisable(GL_TEXTURE_2D);
    // Note that we leave 0 as the last active texture
  }

  if (rendering_mode == GL_RENDER) {
    glLineWidth(1.5);
    glPolygonMode(GL_FRONT, GL_LINE);
    if (Option::render_grid || annotation_type == CONSTRUCTION) {
      for (int m = 0; m < Map::size; ++m) {
        int min_n = Rendering::frustrum_map[m][0];
        int max_n = Rendering::frustrum_map[m][1];
        if (min_n >= max_n) continue;
        draw_triangle_strip(m, min_n, max_n, false, 0.1);
      }
    }
    glPolygonMode(GL_FRONT, GL_FILL);
  }
}

void draw_water() {
  /* set up shiny material */
  GLfloat material_diffuse[] = { 1, 1, 1, 1 };
  GLfloat material_specular[] = { 0.5, 0.5, 0.5, 1 };
  glMaterialfv(GL_FRONT, GL_DIFFUSE, material_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, material_specular);
  glMaterialf(GL_FRONT, GL_SHININESS, 100.0f);

  /* Use the same generic texture, but this time it will be blue */
  if (Option::use_textures) {
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, Texture::rough);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, Texture::polished);
  }
  glPushName(PickName::WATER);
  glNormal3f(0.0, 0.0, 1.0);
  for (int m = 1; m < Map::size-1; ++m) {
      for (int n = Rendering::frustrum_map[m][0];
           n < Rendering::frustrum_map[m][1]; ++n) {

      Rendering::draw_water_triangles(m, n, 1,
                                    Rendering::z_water,
                                    Rendering::r_water, Rendering::g_water,
                                    Rendering::b_water, Rendering::a_water,
                                    annotation_type == CONSTRUCTION? 0.7 : 0.9
                                    );
      Rendering::draw_water_triangles(m, n, 2,
                                    Rendering::z_water,
                                    Rendering::r_water, Rendering::g_water,
                                    Rendering::b_water, Rendering::a_water,
                                    annotation_type == CONSTRUCTION? 0.7 : 0.9
                                    );
      
    }
  }
  
  if (Option::use_textures) {
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glDisable(GL_TEXTURE_2D);
  }

  if (mouseM >= 0 && mouseN >= 0 && Map::water[mouseM][mouseN] > 0.0f) {
    // Highlight the current location
    glColor4f(0.0, 1.0, 1.0, 1.0);
    glNormal3f(0.0, 0.0, 1.0);
    glTexCoord2f(0.0, 0.0);
    Rendering::draw_outline_hexagon(mouseM, mouseN, Rendering::z_water);
  }
  glPopName();
}

// Factory function
GlApp* getGlAppInstance() {
  cout << "Command line options:" << endl
       << "  perspective                   instead of orthographic" << endl
       << "  flat                          instead of gouraud shading" << endl
       << "  fullscreen                    full screen mode" << endl
       << "  no-simulation                 only run the display" << endl
       << "  no-display                    only run the simulation" << endl
       << "  white                         draw gray terrain" << endl
       << "  render-grid                   show the grid all the time" << endl
       << "  no-textures                   use untextured polygons" << endl
       << "  illustration                  ??" << endl
       << endl;
  
  cout << "Controls:" << endl
       << "  Left button: raise land" << endl
       << "  Middle button: lower land" << endl
       << "  Right button: move around on the map" << endl
       << "  PgUp: near-overhead view" << endl
       << "  PgDn: normal view" << endl
       << "  b : blob paths" << endl
       << "  c : canyon rendering options" << endl
       << "  f : flooding" << endl
       << "  g : toggle grid" << endl
       << "  p : toggle perspective" << endl
       << "  t : toggle texturing" << endl
       << "  s : toggle simulation" << endl
       << "  w : toggle white/color" << endl
       << "  v : toggle vectors" << endl
       << "  + : zoom in" << endl
       << "  - : zoom out" << endl
       << "  Ctrl-Q : quit" << endl
       << endl;
  /*
    Thoughts:
    
    Left click: select
    Ctrl-left: add to selection
    Hover: highlight, show info
    Right click: context menu
    Keyboard: scroll view
    Escape: game menu
    P: pause
    M: minimap
  */

  return new GameApp;
}

