#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <iostream>
#include <cmath>
#include <assert.h>
#include <memory>

using namespace std;

#include "main.h"
#include "include_glut.h"

// TODO: use glutWindowStatusFunc to find out when the window is hidden, and disable
// display during that time.  However, simulation should continue.

// TODO: glutSetCursor lets us change the mouse pointer

GlApp* GlApp::singleton = NULL;

namespace Option {
  bool fullscreen = false;
  bool perspective = true;
  bool flat_shading = false;
  double field_of_view = 60.0; // Used only in perspective mode
  float zoom = 60.0; // Used for both perspective and ortho modes
}

void GlApp::event_window_resize(int width, int height) {
  window_width = width;
  window_height = height;
  glutReshapeWindow(width, height);
}

void GlApp::setup_projection() {
  if (Option::perspective) {
    gluPerspective(Option::field_of_view,
                   float(window_width)/float(window_height),
                   1,
                   1000.0);
  } else {
    double multiplier = Option::zoom / 1400.0;
    double isometric_width = window_width * multiplier;
    double isometric_height = window_height * multiplier;
    glOrtho(-isometric_width, isometric_width,
            -isometric_height, isometric_height,
            -100, 200);
  }
}

void GlApp::setup_view() {
  /* make the viewport the size of the window */
  glViewport(0, 0, window_width, window_height);
    
  /* No rotations at first */
  gluLookAt(0.0, 0.0, 100, // Eye position
            0.0, 0.0, 0.0, // Center of vision
            0.0, 1.0, 0.0);// UP angle
}

double get_current_time() {
  timeval tv; gettimeofday(&tv, NULL);
  return double(tv.tv_sec) + double(tv.tv_usec) * 1e-6;
}

struct Timer {
  double last_time;
  double current_time;
  double average_time_per_frame;
  
  Timer(): last_time(0.0), current_time(0.0), average_time_per_frame(0.0) {
  }

  void begin() {
    last_time = get_current_time();
  }

  void end() {
    current_time = get_current_time();
    if ( last_time == 0.0 )
      last_time = current_time;
    else 
      average_time_per_frame = 0.98 * average_time_per_frame
                               + 0.02 * (current_time - last_time);
  }
};

static Timer idle_timer;

void GlApp::idle_handler() {
  idle_timer.end();
  idle_timer.begin();
  glutPostRedisplay();
}

#include "points3.h"
Point3d point_light_source_location(0, 0, 10);

void GlApp::event_paint_window() {
  /* clear the buffer */
  Timer timer;
  timer.begin();

  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  /* set up lights */
  GLfloat amb_light[] = {0.15, 0.15, 0.15, 1.0};
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, amb_light);
  glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 1.0);

  /* Light 0 is movable */
  GLfloat pos0[] = {point_light_source_location.x,
                    point_light_source_location.y,
                    point_light_source_location.z,
                    1.0};
  GLfloat amb0[] = {0.2, 0.2, 0.2, 1.0};
  GLfloat dif0[] = {0.1, 0.1, 0.1, 1.0};
  GLfloat dir0[] = {0.0, 0.0, -1.0, 1.0};
  glLightfv(GL_LIGHT0, GL_AMBIENT, amb0);
  glLightfv(GL_LIGHT0, GL_POSITION, pos0);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, dif0);
  glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, dir0);
  glLightf(GL_LIGHT0, GL_SPOT_CUTOFF, 60.0f);
  glLightf(GL_LIGHT0, GL_SPOT_EXPONENT, 6.0f);

  /* The remaining lights are fixed */
  GLfloat pos1[] = {0, 1000, 1000, 0.0};
  GLfloat dif1[] = {0.5, 0.5, 0.5, 1.0};
  GLfloat spc1[] = {1.0, 1.0, 1.0, 1.0};
  glLightfv(GL_LIGHT1, GL_POSITION, pos1);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, dif1);
  glLightfv(GL_LIGHT1, GL_SPECULAR, spc1);

  GLfloat pos2[] = {-700, -400, 1000, 0.0};
  GLfloat dif2[] = {0.5, 0.5, 0.5, 1.0};
  glLightfv(GL_LIGHT2, GL_POSITION, pos2);
  glLightfv(GL_LIGHT2, GL_DIFFUSE, dif2);

  GLfloat pos3[] = {+700, -400, 1000, 0.0};
  GLfloat dif3[] = {0.5, 0.5, 0.5, 1.0};
  glLightfv(GL_LIGHT3, GL_POSITION, pos3);
  glLightfv(GL_LIGHT3, GL_DIFFUSE, dif3);

  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHT2);
  glEnable(GL_LIGHT3);
  glEnable(GL_LIGHTING);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  /* Annotate with the FPS */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
#if 0
  // TODO: why doesn't this work?  It doesn't seem to be needed in any
  // case... except to set z values, I gues.s
  glOrtho(0, window_width, 0, window_height, -1, 1);
#endif
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glScalef(2.0f / window_width, 2.0f / window_height, 1.0f);
  glTranslatef(-window_width/2.0f, -window_height/2.0f, -0.99f);
#if 0
  // TODO: why doesn't this work?  It doesn't seem to be needed, but
  // it seems like it should be used.
  glViewport(0, window_width, 0, window_height);
#endif
  
  paint_annotation();
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  setup_projection();
  
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  setup_view();
  
  paint_content();
  update_selection();
  ++frame;

  timer.end();
  // Normally __GL_SYNC_TO_VBLANK is stupid and uses up all the CPU
  // while waiting!!  This usleep will instead give it up to the OS.
  
  // usleep(int(idleTimer.currentTime-timer.currentTime));
  
  /* swap draw buffer with display buffer */
  glutSwapBuffers();

}

void GlApp::update_selection() {
}

void GlApp::event_mouse_motion(int x, int y) {
}

void GlApp::event_mouse_button(int button, int state, int x, int y) {
}

void GlApp::event_key(unsigned char key, int x, int y, bool down) {
}

void GlApp::event_special_key(int key, int x, int y, bool down) {
}

void GlApp::paint_annotation() {
  /* TODO:
    From NeHe tutorial 43:

    // A Fairly Straightforward Function That Pushes
// A Projection Matrix That Will Make Object World 
// Coordinates Identical To Window Coordinates.
inline void pushScreenCoordinateMatrix() {
	glPushAttrib(GL_TRANSFORM_BIT);
	GLint   viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(viewport[0],viewport[2],viewport[1],viewport[3]);
	glPopAttrib();
}

// Pops The Projection Matrix Without Changing The Current
// MatrixMode.
inline void pop_projection_matrix() {
	glPushAttrib(GL_TRANSFORM_BIT);
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glPopAttrib();
}

  */
  
  glPushMatrix();
  glLineWidth(2.0);
  glScalef(0.2, 0.2, 0.2);
  glColor3f(1.0, 0.9, 0.0);
  char fps[64];
  sprintf(fps, "%.02f frames/sec", 1.0/idle_timer.average_time_per_frame);
  for (const char* p = fps; *p; ++p) {
    glutStrokeCharacter(GLUT_STROKE_ROMAN, *p);
  }
  glLineWidth(1.0);
  glPopMatrix();
}

void GlApp::paint_content() {
}

// Forwarding functions.  TODO: remove these completely; we shouldn't
// be using member functions for these, because it's not really a
// per-object thing.
namespace {
  
  void _event_window_resize(int width, int height) {
    GlApp::singleton->event_window_resize(width, height);
  }

  void _idle_handler() {
    GlApp::singleton->idle_handler();
  }

  void _event_paint_window() {
    GlApp::singleton->event_paint_window();
  }

  void _event_mouse_motion(int x, int y) {
    GlApp::singleton->event_mouse_motion(x, y);
  }

  void _event_mouse_button(int button, int state, int x, int y) {
    GlApp::singleton->event_mouse_button(button, state, x, y);
  }

  void _event_key_down(unsigned char key, int x, int y) {
    GlApp::singleton->event_key(key, x, y, true);
  }

  void _event_special_key_down(int key, int x, int y) {
    GlApp::singleton->event_special_key(key, x, y, true);
  }

  void _event_key_up(unsigned char key, int x, int y) {
    GlApp::singleton->event_key(key, x, y, false);
  }

  void _event_special_key_up(int key, int x, int y) {
    GlApp::singleton->event_special_key(key, x, y, false);
  }
}

GlApp::GlApp() {
  assert(singleton == NULL);
  singleton = this;
  srand(time(NULL));
  frame = 0;
}

GlApp::~GlApp() {
  cout << "Game Over.  " << frame << " frames." << endl;
  assert(singleton == this);
  singleton = NULL;
}

#if EXTENSIONS_NEEDS_LOADING

/* In Cygwin, these symbols are not in the DLL file.  Instead we have
   to look them up at runtime, using the Windows-specific function
   wglGetProcAddress().  An alternative would be to use
   http://www.levp.de/3d/ */

#include <windows.h>
#include <wingdi.h>

PFNGLMULTITEXCOORD2FARBPROC   ext_glMultiTexCoord2fARB;
PFNGLACTIVETEXTUREARBPROC     ext_glActiveTextureARB;
PFNGLCLIENTACTIVETEXTUREARBPROC     ext_glClientActiveTextureARB;
PFNGLTEXIMAGE3DPROC           ext_glTexImage3D;

void LoadExtensions() {
  ext_glActiveTextureARB =
    (PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress("glActiveTextureARB");
  ext_glClientActiveTextureARB =
    (PFNGLCLIENTACTIVETEXTUREARBPROC)wglGetProcAddress("glClientActiveTextureARB");
  ext_glMultiTexCoord2fARB =
    (PFNGLMULTITEXCOORD2FARBPROC)wglGetProcAddress("glMultiTexCoord2fARB");
  ext_glTexImage3D = (PFNGLTEXIMAGE3DPROC)wglGetProcAddress("glTexImage3D");

  if (ext_glMultiTexCoord2fARB == NULL
      || ext_glClientActiveTextureARB == NULL
      || ext_glMultiTexCoord2fARB == NULL) {
    cerr << "This program requires the OpenGL multitexturing extension." << endl;
    exit(1);
  }
  
  if (ext_glTexImage3D == NULL) {
    cerr << "This program requires the OpenGL 3D texture extension." << endl;
    exit(1);
  }
}

#endif

void GlApp::setup_gl() {
  /* setup and create the window */
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
  glutCreateWindow(window_name());
  if (Option::fullscreen) glutFullScreen();

#if EXTENSIONS_NEEDS_LOADING
  LoadExtensions();
#endif

  cerr << "OpenGL information:" << endl;
  cerr << "  Vendor:   " << reinterpret_cast<const char *>(glGetString(GL_VENDOR)) << endl;
  cerr << "  Renderer: " << reinterpret_cast<const char *>(glGetString(GL_RENDERER)) << endl;
  cerr << "  Version:  " << reinterpret_cast<const char *>(glGetString(GL_VERSION)) << endl;
  
  /* setup stuff in here */
  // glEnable(GL_POINT_SMOOTH);
  // glEnable(GL_LINE_SMOOTH);
  // glEnable(GL_POLYGON_SMOOTH);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_ALPHA_TEST);
  glEnable(GL_BLEND);
  if (Option::flat_shading) {
    glShadeModel(GL_FLAT);
  }
  // glDisable(GL_DITHER);

  glutIgnoreKeyRepeat(true);
  glutSetCursor(GLUT_CURSOR_FULL_CROSSHAIR);
  
  glutDisplayFunc(_event_paint_window);
  glutReshapeFunc(_event_window_resize);
  glutIdleFunc(_idle_handler);
  glutPassiveMotionFunc(_event_mouse_motion);
  glutMotionFunc(_event_mouse_motion);
  glutMouseFunc(_event_mouse_button);
  glutKeyboardFunc(_event_key_down);
  glutSpecialFunc(_event_special_key_down);
  glutKeyboardUpFunc(_event_key_up);
  glutSpecialUpFunc(_event_special_key_up);
}

bool GlApp::set_flag(const string& flag) {
  if (flag == "perspective") Option::perspective = true;
  else if (flag == "fullscreen") Option::fullscreen = true;
  else if (flag == "flat") Option::flat_shading = true;
  else return false;

  return true;
}

void GlApp::go() {
  glutMainLoop();
}

int main(int argc, char* argv[]) {
  auto_ptr<GlApp> app(getGlAppInstance());
  pair<int, int> window_size = app->window_size();
  glutInitWindowSize(window_size.first, window_size.second);
  glutInit(&argc, argv);
  for (int i = 1; i < argc; ++i) {
    if (!app->set_flag(argv[i])) {
      cerr << "Unrecognized option: " << argv[i] << endl;
    }
  }
  
  app->setup_gl();
  app->go();
}

