#include <cmath>
#include <iostream>
#include <vector>

#include "main.h"
#include "points.h"

struct Spline2dSegment {
  Vector2d a, b, c; Point2d d;

  Spline2dSegment(const Point2d& p, Vector2d dp,
                  const Point2d& q, Vector2d dq) {
    dp = normalize(dp);
    dq = normalize(dq);
    double scale1 = 2 * dot(q-p, dp);
    double scale2 = 2 * dot(q-p, dq);

    setControls(p, dp * scale1, q, dq * scale2);
  }

  void setControls(const Point2d& p, Vector2d dp,
                   const Point2d& q, Vector2d dq) {
    // Assume dp, dq are normalized
    a = Vector2d(2*p.x + dp.x - 2*q.x + dq.x,
                 2*p.y + dp.y - 2*q.y + dq.y);
    b = Vector2d(-3*p.x - 2*dp.x + 3*q.x - dq.x,
                 -3*p.y - 2*dp.y + 3*q.y - dq.y);
    c = dp;
    d = p;
  }
  
  Point2d evaluate(double t) {
    return a*(t*t*t) + b*(t*t) + c*t + d;
  }

  Vector2d slope(double t) {
    return 3*a*(t*t) + 2*b*t + c;
  }

  double curvature(double t) {
    Vector2d n = slope(t);
    Vector2d m = 6*a*t + 2*b;
    double nlen = n.length();
    return nlen*nlen*nlen / (n.x*m.y - n.y*m.x);
  }
};  
  
struct Spline2d {
  vector<Point2d> p;
  vector<Vector2d> v;
  vector<Spline2dSegment> segments;

  int size() const { return p.size(); }

  void set_point(int index, const Point2d& new_point) {
    p[index] = new_point;
    if (index > 0) {
      segments[index-1] = Spline2dSegment(p[index-1], v[index-1],
                                          p[index], v[index]);
    }
    if (index < p.size()-1) {
      segments[index] = Spline2dSegment(p[index], v[index],
                                        p[index+1], v[index+1]);
    }
  }

  void set_vector(int index, const Point2d& new_endpoint) {
    Vector2d new_v = new_endpoint - p[index];
    if (index > 0) {
      segments[index-1] = Spline2dSegment(p[index-1], v[index-1], p[index], new_v);
    }
    if (index < p.size()-1) {
      segments[index] = Spline2dSegment(p[index], new_v, p[index+1], v[index+1]);
    }
    v[index] = new_v;
  }
  
  void append(const Point2d& new_point, const Vector2d& new_vector) {
    if (!p.empty()) {
      segments.push_back(Spline2dSegment(p.back(), v.back(),
                                         new_point, new_vector));
    }
    p.push_back(new_point);
    v.push_back(new_vector);
  }
  
  pair<double, Spline2dSegment> get_segment(double t) const {
    t *= segments.size();
    int segment_number = int(t);
    if (segment_number < 0)
      segment_number = 0;
    if (segment_number >= segments.size())
      segment_number = segments.size() - 1;
    t = t - segment_number;
    return make_pair(t, segments[segment_number]);
  }
  
  Point2d evaluate(double t) const {
    pair<double, Spline2dSegment> r = get_segment(t);
    return r.second.evaluate(r.first);
  }

  Vector2d slope(double t) const {
    pair<double, Spline2dSegment> r = get_segment(t);
    return r.second.slope(r.first);
  }

  double curvature(double t) const {
    pair<double, Spline2dSegment> r = get_segment(t);
    return r.second.curvature(r.first);
  }
};

int rawX = 0;
int rawY = 0;
const double WIDTH = 5.0;

void draw(const Spline2d& spline) {
  double alpha = 1.0;
 
  // Layer 0.0 is the road surface 
  glNormal3f(0.0, 0.0, 1.0);
  glBegin(GL_TRIANGLE_STRIP);
  const int N = 100;
  double L = 0.0; // Our texture coordinate, measured in pixel distance
  Point2d prev_p = spline.evaluate(0.0);
  for (int t = 0; t <= N; ++t) {
    Point2d p = spline.evaluate(1.0*t/N);
    Vector2d v = WIDTH * normalize(spline.slope(1.0*t/N));
    double curvature = abs(spline.curvature(1.0*t/N));
    if (curvature <= WIDTH || dot(v, spline.slope(1.0*t/N+0.01)) < 0)
      glColor4f(1.0, 0.0, 0.0, alpha);
    else if (curvature <= WIDTH*1.5)
      glColor4f(1.0, 1.0, 0.0, alpha);
    else
      glColor4f(0.7, 0.7, 0.7, alpha);
    L += (p-prev_p).length(); prev_p = p;
    glTexCoord2f(L, 0.0); glVertex3f(p.x + v.y, p.y - v.x, 0.0);
    glTexCoord2f(L, 1.0); glVertex3f(p.x - v.y, p.y + v.x, 0.0);
  }
  glEnd();
}

int mouseDown = 0;
int which_point = -1;
bool point_mode = true;
const int EDITOR_V_DIST = 10;

Spline2d spline;

const int TILE_MAP_SIZE = 128;
double height[TILE_MAP_SIZE][TILE_MAP_SIZE];
double water[TILE_MAP_SIZE][TILE_MAP_SIZE];

Point2d centerOfView(0.0, 0.0);
Point2d centerOfViewTarget(TILE_MAP_SIZE*0.4, TILE_MAP_SIZE*0.4);
Vector2d angleOfView(0.0, 1.0);
GLdouble mvmatrix[16], projmatrix[16];
GLint viewport[4];

Point2d screenToWorld(double screenX, double screenY, double worldZ) {
  double wx0, wy0, wz0;
  gluUnProject(screenX, screenY, 0.0,
               mvmatrix, projmatrix, viewport, &wx0, &wy0, &wz0);
  double wx1, wy1, wz1;
  gluUnProject(screenX, screenY, 1.0,
               mvmatrix, projmatrix, viewport, &wx1, &wy1, &wz1);
  
  // Compute a position on the line (wx0,wy0,wz0)-(wx1,wy1,wz1) where z = worldZ
  double depth = (worldZ-wz0) / (wz1-wz0);
  double x = wx0 + (wx1-wx0)*depth;
  double y = wy0 + (wy1-wy0)*depth;
  return Point2d(x, y);
}

Point2d computeMousePosition(double worldZ) {
  int realY = viewport[3] - rawY - 1;
  return screenToWorld(rawX, realY, worldZ);
}

struct SplineApp: GlApp {

  virtual void setupView() {
    /* Change view if the mouse button is down */
    if (mouseDown == 3) {
      centerOfViewTarget = computeMousePosition(0.0);
    }
    centerOfView = centerOfView + 0.03*(centerOfViewTarget - centerOfView);
    Vector2d n = centerOfViewTarget - centerOfView;
    if (n.length() > 0.1) {
      angleOfView = normalize(angleOfView + 0.0008*n);
    }

    /* Set up matrices */
    // TODO: as we're rotating the camera, also change the up angle so that it looks
    // like we're turning like a plane/helicopter
    glLoadIdentity();
    Point2d eye = centerOfView + -25*angleOfView;
    gluLookAt(eye.x, eye.y, 15, // Eye position
              centerOfView.x, centerOfView.y, 0.0, // Center of vision
              0.0, 0.0, 1.0);// UP angle
    glGetIntegerv(GL_VIEWPORT, viewport);
    glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix);
    glGetDoublev(GL_PROJECTION_MATRIX, projmatrix);
  }

  /* Reaction-diffusion pattern */
  void reactionDiffusion() {
    const int W = 64, H = 64;
    static double A[W][H] = {0.5};
    static double B[W][H] = {0.5};
    static double s = 0.2/16.0;
    static int frame = 0;
    if (++frame == 1) {
      for (int i = 0; i < W; ++i) {
        for (int j = 0; j < H; ++j) {
          A[i][j] = 4 + 0.01*(random()%200-random()%200);
          B[i][j] = 4 + 0.01*(random()%200-random()%200);
        }
      }
    }

    double da[W][H], db[W][H];
    for (int i = 0; i < W; ++i) {
      for (int j = 0; j < H; ++j) {
        double C = 12 + 0.1*((3*i+5*j)%300);
        da[i][j] = s*(16-A[i][j]*B[i][j])
                   + 0.5*0.25*(A[(i+1)%W][j] + A[(i-1+W)%W][j]
                               + A[i][(j+1)%H] + A[i][(j-1+H)%H]
                               - 4*A[i][j]);
        db[i][j] = s*(A[i][j]*B[i][j] - B[i][j] - C)
                   + 0.5*0.0625*(B[(i+1)%W][j] + B[(i-1+W)%W][j]
                                 + B[i][(j+1)%H] + B[i][(j-1+H)%H]
                                 - 4*B[i][j]);
      }
    }

    A[W/2][H/4] = 4;
    B[W/2][H/4] = 0;
    A[W/4][H/2] = 0;
    B[W/4][H/2] = 4;
    double speed = 0.01;
    for (int i = 0; i < W; ++i) {
      for (int j = 0; j < H; ++j) {
        A[i][j] = max(min(8.0, A[i][j]+speed*da[i][j]), 0.0);
        B[i][j] = max(min(8.0, B[i][j]+speed*db[i][j]), 0.0);
        glColor3f(0.12*A[i][j], 0.0, 0.12*B[i][j]);
        glBegin(GL_POLYGON);
        glVertex3f(i, 20, j);
        glVertex3f(i+1, 20, j);
        glVertex3f(i+1, 20, j+1);
        glVertex3f(i, 20, j+1);
        glEnd();
      }
    }
  }

  inline double z(double x, double y) {
    #if 0
    double _z = x*0.01-y*0.001;
    if (x > 70 && x < 100) _z += 0.01*y*(x-70);
    if (y > 100 && y < 140) _z -= 0.01*x*(y-100);
    return _z;
    #endif
    double dx = 0.01*(x-110), dy = 0.01*(y-110);
    return 8.0*(sin(x*0.1)+cos(y*0.1))/(dx*dx+dy*dy+1);
  }

  /** temp: slope by using lines crossing (a0,a1) crossing (b0,b1)  */
  void slope(Point2d a0, Point2d a1,
             Point2d b0, Point2d b1) {
    double dx0 = a1.x-a0.x, dx1 = b1.x-b0.x;
    double dy0 = a1.y-a0.y, dy1 = b1.y-b0.y;
    double dz0 = z(a1.x, a1.y) - z(a0.x, a0.y);
    double dz1 = z(b1.x, b1.y) - z(b0.x, b0.y);
    double normX = dy0*dz1 - dz0*dy1;
    double normY = dz0*dx1 - dx0*dz1;
    double normZ = dx0*dy1 - dy0*dx1;
    double n = sqrt(normX*normX + normY+normY + normZ*normZ);
    normX /= n;
    normY /= n;
    normZ /= n;
    glNormal3f(normX, normY, normZ);
  }
  
  virtual void paintContent() {
    const double t_size = 2.5; // triangle size
    double xWidth = t_size*sqrt(0.75);
    double yHeight = t_size;
    
    // TEMP: set up texture; move this outside of paint
    static unsigned int tex_num = 0;
    const int NUM_TEXTURES = 4;
    static unsigned int tex_anim[NUM_TEXTURES];
    if (tex_num == 0) {
      const int W = 128, H = 128;
      static unsigned char texture[W*H*4];
      for (int i = 0; i < W; ++i) {
        for (int j = 0; j < H; ++j) {
          bool road_edge = i==2 || i==W-3;
          bool center_divider = (i==W/2+1) || (i==W/2-1);
          bool lane = !center_divider
                      && !road_edge
                      && (i != W/2)
                      && i > 2
                      && ((i*8) % W == 0)
                      && j < H/3;
          int alpha = 255;
          if (i < 1 || i > W-2) alpha = 0;

          texture[i*W*4+j*4] = (road_edge||center_divider||lane)? 255:128;
          texture[i*W*4+j*4+1] = (road_edge||center_divider||lane)? 255:128;
          texture[i*W*4+j*4+2] = (road_edge||lane)? 255:128;
          texture[i*W*4+j*4+3] = alpha;
        }
      }

      glGenTextures(1, &tex_num);
      glBindTexture(GL_TEXTURE_2D, tex_num); 
      glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
      glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, W, H, 0, GL_RGBA, GL_UNSIGNED_BYTE,
                   texture);

      cout << "Setting up textures." << endl;
      
      for (int k = 0; k < NUM_TEXTURES; ++k) {
        for (int i = 0; i < W; ++i) {
          for (int j = 0; j < H; ++j) {
            int r = 96 + random()%64 - random()%64;
            int g = 128 + random()%64 + random()%64;
            int b = 64 + random()%32 - random()%32;
            int a = 255;

            texture[i*W*4+j*4] = r;
            texture[i*W*4+j*4+1] = g;
            texture[i*W*4+j*4+2] = b;
            texture[i*W*4+j*4+3] = a;
          }
        }
        glGenTextures(1, &tex_anim[k]);
        glBindTexture(GL_TEXTURE_2D, tex_anim[k]);
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
                     W, H, 0, GL_RGBA, GL_UNSIGNED_BYTE,
                     texture);
      }

      for (int i = 0; i < TILE_MAP_SIZE; ++i) {
        for (int j = 0; j < TILE_MAP_SIZE; ++j) {
          height[i][j] = z(i*xWidth, j*yHeight + (i&1)*0.5*yHeight);
          water[i][j] = 0.0;
        }
      }
    }
  
    // straight and a quarter on each end is curved.  If instead
    // |slope|=sqrt(2)*d, the entire road is curved.  This isn't quite
    // right; it only works when the angle between the slopes and the
    // straight line distance is roughly 45 degrees.  :(
    if (spline.size() == 0) {
      spline.append(Point2d(0, 0), Vector2d(78, 0));
      spline.append(Point2d(40, 40), Vector2d(0, 78));
      spline.append(Point2d(80, 80), Vector2d(78, 0));
      spline.append(Point2d(100, 100), Vector2d(39, 0));
    }

    // reactionDiffusion();

    // Figure out how much of the map we actually need to draw.
    Point2d minP, maxP;
    {
      const int gutter = 30;
      Point2d a = screenToWorld(viewport[0]-gutter, viewport[1]-gutter, 0.0);
      Point2d b = screenToWorld(viewport[2]+gutter, viewport[3]+gutter, 0.0);
      Point2d c = screenToWorld(viewport[0]-gutter, viewport[3]+gutter, 0.0);
      Point2d d = screenToWorld(viewport[2]+gutter, viewport[1]-gutter, 0.0);
      minP = a;
      minP.x = min(minP.x, b.x); minP.y = min(minP.y, b.y);
      minP.x = min(minP.x, c.x); minP.y = min(minP.y, c.y);
      minP.x = min(minP.x, d.x); minP.y = min(minP.y, d.y);
      maxP = a;
      maxP.x = max(maxP.x, b.x); maxP.y = max(maxP.y, b.y);
      maxP.x = max(maxP.x, c.x); maxP.y = max(maxP.y, c.y);
      maxP.x = max(maxP.x, d.x); maxP.y = max(maxP.y, d.y);
    }
    
    /* Draw a triangular grid */
    const double T = 0.05; // texture scale
    const double sqrt_of_3_4 = sqrt(0.75);
    glEnable(GL_TEXTURE_2D);
#define P(i,j) Point2d((i)*sqrt_of_3_4*t_size, ((j)+0.5*((i)&1))*t_size)
    for (int i = 1; i < TILE_MAP_SIZE-1; ++i) {
      glBindTexture(GL_TEXTURE_2D, tex_anim[i%NUM_TEXTURES]);
      glBegin(GL_TRIANGLE_STRIP);
      for (int j = 1; j < TILE_MAP_SIZE-1; ++j) {
        Point2d p0 = P(i, j);
        Point2d p1 = P(i+1, j);

        if (p0.x < minP.x || p0.x > maxP.x || p0.y < minP.y || p0.y > maxP.y )
          continue;
        
        glColor3f(0.8, 0.8, 0.6);
        glTexCoord2f(p0.x*T, p0.y*T);
        slope(P(i-1,j), P(i+1,j), P(i,j-1), P(i,j+1));
        glVertex3f(p0.x, p0.y, -1.0+z(p0.x, p0.y));
        glColor3f(0.8, 0.8, 0.6);
        glTexCoord2f(p1.x*T, p1.y*T);
        slope(P(i-1,j+1), P(i+1,j+1), P(i,j+1-(i&1)), P(i,j+3-(i&1)));
        glVertex3f(p1.x, p1.y, -1.0+z(p1.x, p1.y));
      }
      glEnd();
    }
    glDisable(GL_TEXTURE_2D);
    
    /* I want water to write over land, even though they are at the same z value */
    glDepthFunc(GL_LEQUAL);

    /* Everything else won't have normal normals */
    glNormal3f(0.0, 0.0, 1.0);

    /* The current mouse location, converted to world coordinates */
    Point2d mouse = computeMousePosition(0.0);

    /* Draw water surface.  TODO: for proper display, hexagons in
       *front* should be drawn first.  This will allow transparency
       and z-order to work out better.
    */
#if 1
    for (int i = 0; i < TILE_MAP_SIZE; ++i) {
      for (int j = 0; j < TILE_MAP_SIZE; ++j) {
        double x0 = i*xWidth;
        double y0 = j*yHeight + (i&1)*0.5*yHeight;
        double zbase = -1.0+height[i][j];
        double z0 = water[i][j]+zbase;
        if (abs(x0 - mouse.x) < 3 && abs(y0 - mouse.y) < 3)
          water[i][j] += 0.2;
      
        if (water[i][j] > 0.001) {
          water[i][j] *= 0.995;

          if (x0 < minP.x || x0 > maxP.x || y0 < minP.y || y0 > maxP.y)
            continue; // TODO: this continue is in a bad place ..
          
          /* Draw a hexagon at i,j */
          glColor4f(0.1, 0.8/(1.0+water[i][j]), 1.0, min(0.1+10.0*water[i][j], 0.5));
          double dx[6] = {2, 1, -1, -2, -1, 1};
          double dy[6] = {0, 1, 1, 0, -1, -1};
#define V(x,y) glVertex3f((x),(y),z0)
#define V2(x,y) glVertex3f((x),(y),zbase-3)
          
          glBegin(GL_TRIANGLE_FAN);
          V(x0,y0);
          for (int k = 0; k <= 6; ++k) {
            V(x0 + xWidth*2/6*dx[k%6], y0 + yHeight*0.5*dy[k%6]);
          }
          glEnd();
        }
      }
    }
#endif

    /* Draw water sides (doesn't make any sense but until we have slanted
       water surfaces, this will have to do */
#if 1
    for (int i = 0; i < TILE_MAP_SIZE; ++i) {
      for (int j = 0; j < TILE_MAP_SIZE; ++j) {
        if (water[i][j] > 0.001) {
          double x0 = i*xWidth;
          double y0 = j*yHeight + (i&1)*0.5*yHeight;
          double zbase = -1.0+height[i][j];
          double z0 = water[i][j]+zbase;
          
          double dx[6] = {2, 1, -1, -2, -1, 1};
          double dy[6] = {0, 1, 1, 0, -1, -1};
          
          glColor4f(0.3, 0.3, 0.3, 1.0);
          for (int k = 0; k < 6; ++k) {
            glBegin(GL_QUADS);
            V(x0 + xWidth*2/6*dx[k%6], y0 + yHeight*0.5*dy[k%6]);
            V(x0 + xWidth*2/6*dx[(k+1)%6], y0 + yHeight*0.5*dy[(k+1)%6]);
            V2(x0 + xWidth*2/6*dx[(k+1)%6], y0 + yHeight*0.5*dy[(k+1)%6]);
            V2(x0 + xWidth*2/6*dx[k%6], y0 + yHeight*0.5*dy[k%6]);
            glEnd();
          }
        }
      }
    }
#endif

    #if 1
    /* Change tile map */
    const int NDIR = 8;
    static const int dir[8][2] = {
      {-1, 0}, {0, -1}, {+1, 0}, {0, +1},
      {-1, -1}, {-1, +1}, {+1, +1}, {+1, -1},
    };
    static int lastdir = 0;
    lastdir = (lastdir+1) % NDIR;
    
    double dwater[TILE_MAP_SIZE][TILE_MAP_SIZE] = {0.0};
    for (int i = 1; i < TILE_MAP_SIZE-1; ++i) {
      for (int j = 1; j < TILE_MAP_SIZE-1; ++j) {
        int best_k = -1;
        double best_w = 0;
        for (int k = lastdir; k < lastdir+NDIR; ++k) {
          int di = dir[k%NDIR][0], dj = dir[k%NDIR][1];
          if (di != 0 || dj != 0) {
            double dh = water[i][j]+height[i][j]
                        - (water[i+di][j+dj]+height[i+di][j+dj]);
            if (i > 0 && dh > 0 && water[i][j] > 0) {
              // Flow from one to another
              double w = min(water[i][j], dh*0.2);
              if (w >= best_w) {
                best_k = k%NDIR;
                best_w = w;
              }
            }
          }
        }
        if (best_k != -1) {
          int di = dir[best_k][0], dj = dir[best_k][1];
          dwater[i][j] -= best_w;
          dwater[i+di][j+dj] += best_w;
        }
      }
    }
    for (int i = 1; i < TILE_MAP_SIZE-1; ++i) {
      for (int j = 1; j < TILE_MAP_SIZE-1; ++j) {
        water[i][j] += dwater[i][j];
      }
    }
    
    #endif

    #if 0
    /* Test texture */
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, tex_num);
  
    /* Draw the road */
    draw(spline);
    
    glDisable(GL_TEXTURE_2D);
    #endif

    #if 1
    /* Draw mouse pointer */
    {
      Point2d mouse = computeMousePosition(0.0);
      glColor4f(1.0, 1.0, 0.0, 0.6);
      glBegin(GL_POLYGON);
      glVertex3f(mouse.x-2, mouse.y-2, 0.2);
      glVertex3f(mouse.x+2, mouse.y-2, 0.2);
      glVertex3f(mouse.x+2, mouse.y+2, 0.2);
      glVertex3f(mouse.x-2, mouse.y+2, 0.2);
      glEnd();
      glColor3f(1.0, 0.0, 0.0);
      glEnable(GL_LINE_SMOOTH);
      glBegin(GL_LINE_STRIP);
      glVertex3f(mouse.x-2, mouse.y-2, 0.2);
      glVertex3f(mouse.x+2, mouse.y-2, 0.2);
      glVertex3f(mouse.x+2, mouse.y+2, 0.2);
      glVertex3f(mouse.x-2, mouse.y+2, 0.2);
      glVertex3f(mouse.x-2, mouse.y-2, 0.2);
      glVertex3f(mouse.x, mouse.y, 7);
      glVertex3f(mouse.x-2, mouse.y+2, 0.2);
      glVertex3f(mouse.x, mouse.y, 7);
      glVertex3f(mouse.x+2, mouse.y-2, 0.2);
      glVertex3f(mouse.x, mouse.y, 7);
      glVertex3f(mouse.x+2, mouse.y+2, 0.2);
      glEnd();
      glDisable(GL_LINE_SMOOTH);
    }
    #endif
    
    /* Draw grids at mouse pointer */
#if 1
    glLineWidth(0.5);
    const int BOX = 3;
    for (int i = -20; i <= 20; i+=BOX) {
      for (int j = -10; j <= 10; j+=BOX) {
        glColor4f(1.0, 1.0, 1.0, 0.3);
        glBegin(GL_LINE_STRIP);
        glVertex3f(mouse.x + i, mouse.y, j);
        glVertex3f(mouse.x + i + BOX, mouse.y, j);
        glVertex3f(mouse.x + i + BOX, mouse.y, j + BOX);
        glVertex3f(mouse.x + i, mouse.y, j + BOX);
        glVertex3f(mouse.x + i, mouse.y, j);
        glEnd();
        glBegin(GL_LINE_STRIP);
        glVertex3f(mouse.x, mouse.y + i, j);
        glVertex3f(mouse.x, mouse.y + i + BOX, j);
        glVertex3f(mouse.x, mouse.y + i + BOX, j + BOX);
        glVertex3f(mouse.x, mouse.y + i, j + BOX);
        glVertex3f(mouse.x, mouse.y + i, j);
        glEnd();
      }
    }
    for (int i = -20; i <= 20; i+=BOX) {
      for (int j = -10; j <= 10; j+=BOX) {
        glColor4f(0.1, 0.1, 0.1, 0.2);
        glBegin(GL_POLYGON);
        glVertex3f(mouse.x + i, mouse.y, j);
        glVertex3f(mouse.x + i + BOX, mouse.y, j);
        glVertex3f(mouse.x + i + BOX, mouse.y, j + BOX);
        glVertex3f(mouse.x + i, mouse.y, j + BOX);
        glEnd();
        glBegin(GL_POLYGON);
        glVertex3f(mouse.x, mouse.y + i, j);
        glVertex3f(mouse.x, mouse.y + i + BOX, j);
        glVertex3f(mouse.x, mouse.y + i + BOX, j + BOX);
        glVertex3f(mouse.x, mouse.y + i, j + BOX);
        glEnd();
      }
    }
#endif

    #if 0
    /* Draw the control points and control vectors */
    glPointSize(8.0);
    glLineWidth(1.5);
    glEnable(GL_LINE_SMOOTH);
    for (int i = 0; i < spline.size(); ++i) {
      Point2d endpoint1 = spline.p[i] + EDITOR_V_DIST*normalize(spline.v[i]);
      Point2d endpoint2 = spline.p[i] - EDITOR_V_DIST*normalize(spline.v[i]);
      if (i == which_point && !point_mode) {
        glColor4f(1.0, 0.0, 0.0, 1.0);
      } else {
        glColor4f(0.0, 0.0, 1.0, 1.0);
      }
      glBegin(GL_LINES);
      /* Orientation */
      glVertex3f(endpoint1.x, endpoint1.y, 1.0);
      glVertex3f(endpoint2.x, endpoint2.y, 1.0);
      /* Control point to ground */
      glVertex3f(spline.p[i].x, spline.p[i].y, 1.0);
      glVertex3f(spline.p[i].x, spline.p[i].y, 0.0);
      glEnd();
      glBegin(GL_POINTS);
      glVertex3f(endpoint1.x, endpoint1.y, 1.0);
      glVertex3f(endpoint2.x, endpoint2.y, 1.0);
      if (i == which_point && point_mode) {
        glColor4f(1.0, 0.0, 0.0, 1.0);
      } else {
        glColor4f(0.0, 0.0, 1.0, 1.0);
      }
      glVertex3f(spline.p[i].x, spline.p[i].y, 1.0);
      glEnd();
    }
    glDisable(GL_LINE_SMOOTH);
    #endif
  }

  virtual void mouseMotionHandler(int x, int y) {
    rawX = x;
    rawY = y;
    /* Control points are located at z=1 */
    Point2d p = computeMousePosition(1.0);
    if (mouseDown == 1) {
      if (which_point >= 0) {
        if (point_mode) {
          spline.set_point(which_point, Point2d(p.x, p.y));
        } else {
          spline.set_vector(which_point, Point2d(p.x, p.y));
        }
      }
    }
  }

  virtual void mouseButtonHandler(int button, int state, int x, int y) {
    mouseMotionHandler(x, y);
    if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
      // Mouse down event; find the closest control point (which are drawn at z=1)
      Point2d p = computeMousePosition(1.0);
      mouseDown = 1;
      int j = -1;
      double best_distance = 10.0;
      for (int i = 0; i < spline.size(); ++i) {
        Point2d control_point = spline.p[i];
        Point2d endpoint1 = spline.p[i] + EDITOR_V_DIST*normalize(spline.v[i]);
        Point2d endpoint2 = spline.p[i] - EDITOR_V_DIST*normalize(spline.v[i]);
        double d0 = dist(control_point, p);
        if ( d0 < best_distance ) {
          best_distance = d0;
          j = i;
          point_mode = true;
        }
        double d1 = dist(endpoint1, p);
        if ( d1 < best_distance ) {
          best_distance = d1;
          j = i;
          point_mode = false;
        }
        double d2 = dist(endpoint2, p);
        if ( d2 < best_distance ) {
          best_distance = d2;
          j = i;
          point_mode = false;
        }
      }
      which_point = j;
      cout << "Grabbing mouse " << button << " == control point " << which_point << endl;
      mouseMotionHandler(x, y);
    } else if (button == GLUT_LEFT_BUTTON && state == GLUT_UP) {
      cout << "Releasing mouse " << button << endl;
      which_point = -1;
      mouseDown = 0;
    } else if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN) {
      mouseDown = 3;
      mouseMotionHandler(x, y);
    } else if (button == GLUT_RIGHT_BUTTON && state == GLUT_UP) {
      mouseDown = 0;
    }
  }
};

// Factory function
GlApp* getGlAppInstance() {
  return new SplineApp;
}
